import re
import importlib
import importlib.resources as importlib_resources
from types import SimpleNamespace

import numpy as np
import pytest
from PIL import Image

import topovec.marimo as tvm
camera_viewer_module = importlib.import_module("topovec.marimo.camera_viewer")
from topovec.core.property import BooleanProperty, ListProperty, NumericProperty
from topovec.core import System, UpdateMask
from topovec.marimo.camera_viewer import (
    DirectorFieldView,
    RenderSession,
    _camera_viewport_widget_class,
    _effective_numeric_step,
    _quantize_numeric_value,
    camera_state_signature,
    inspect as inspect_layout,
    inspect_view,
    restore_camera,
    serialize_camera,
)
from topovec.mgl.camera import Camera, UnitCamera


class DummySystem:
    def __init__(self, dim=3):
        self.dim = dim


class DummySceneBase:
    def __init__(self, mglctx=None, system=None, camera=None):
        self._mglctx = mglctx
        self._system = DummySystem() if system is None else system
        self.camera = Camera() if camera is None else camera
        self._cameras = {str(self.camera_slot_key): self.camera}
        self.uploaded_state = None
        self.list_properties_calls = 0
        self.activate_camera_slot()

    @property
    def system(self):
        return self._system

    @property
    def camera_slot_key(self):
        return self.camera_profile

    @property
    def camera_profile(self):
        return "general"

    def _make_default_camera(self, slot):
        return Camera()

    def activate_camera_slot(self):
        slot = str(self.camera_slot_key)
        if slot not in self._cameras:
            self._cameras[slot] = self._make_default_camera(slot)
        self.camera = self._cameras[slot]
        return self.camera

    @classmethod
    def supports_system(cls, system):
        return True

    def upload(self, state=None):
        self.uploaded_state = state


class DummyScene(DummySceneBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.flag = False
        self.level = 0.5
        self._follow_view = False
        self._mode = "alpha"
        self._min_level = 0.0
        self._max_level = 1.0
        self._axis = 2
        self.axis = 2

    @property
    def camera_profile(self):
        return f"layer-{('x', 'y', 'z')[getattr(self, '_axis', 2)]}"

    def _make_default_camera(self, slot):
        camera = Camera()
        if slot == "layer-x":
            camera.set_basis(depth=[1, 0, 0], up=[0, 0, 1])
        elif slot == "layer-y":
            camera.set_basis(depth=[0, 1, 0], up=[1, 0, 0])
        else:
            camera.set_basis(depth=[0, 0, 1], up=[1, 0, 0])
        return camera

    def get_mode(self):
        return self._mode

    def set_mode(self, value):
        self._mode = value
        if value == "alpha":
            self._min_level = 0.0
            self._max_level = 1.0
        else:
            self._min_level = -2.0
            self._max_level = 2.0
        self.level = float(np.clip(self.level, self._min_level, self._max_level))
        return UpdateMask.REDRAW | UpdateMask.REBUILD_PROPERTIES
    mode = property(get_mode, set_mode)

    def get_axis(self):
        return self._axis

    def set_axis(self, value):
        self._axis = int(value)
        return (
            UpdateMask.REDRAW
            | UpdateMask.REBUILD_PROPERTIES
            | UpdateMask.CAMERA_SLOT_CHANGED
        )
    axis = property(get_axis, set_axis)

    def get_follow_view(self):
        return self._follow_view

    def set_follow_view(self, value):
        self._follow_view = bool(value)
        if self._follow_view:
            self.camera.set_basis(depth=[1, 1, 0], up=[0, 0, 1])
        else:
            self.camera.set_basis(depth=[0, 0, 1], up=[1, 0, 0])
        return UpdateMask.REDRAW
    follow_view = property(get_follow_view, set_follow_view)

    def list_properties(self):
        self.list_properties_calls += 1
        return {
            "Render": [
                ListProperty(
                    self,
                    "mode",
                    title="Mode",
                    values=["alpha", "beta"],
                    reset=True,
                ),
                NumericProperty(
                    self,
                    "level",
                    title="Level",
                    min=self._min_level,
                    max=self._max_level,
                    count=5,
                    param_key="shared.level",
                ),
                BooleanProperty(
                    self,
                    "flag",
                    title="Flag",
                    param_key="shared.flag",
                ),
                NumericProperty(
                    self,
                    "axis",
                    title="Axis",
                    min=0,
                    max=2,
                    count=3,
                    reset=True,
                    param_key="layer.axis",
                ),
                BooleanProperty(self, "follow_view", title="Follow view"),
            ]
        }


class DummyGeneralScene(DummySceneBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.level = 0.25
        self.flag = False

    def list_properties(self):
        self.list_properties_calls += 1
        return {
            "General": [
                NumericProperty(
                    self,
                    "level",
                    title="Level",
                    min=0.0,
                    max=1.0,
                    count=5,
                    param_key="shared.level",
                ),
                BooleanProperty(
                    self,
                    "flag",
                    title="Flag",
                    param_key="shared.flag",
                )
            ]
        }


class DummyBlochScene(DummySceneBase):
    @property
    def camera_profile(self):
        return "bloch"

    def _make_default_camera(self, slot):
        camera = Camera()
        camera.set_basis(depth=[0, 1, 1], up=[1, 0, 0])
        return camera

    def list_properties(self):
        self.list_properties_calls += 1
        return {"Ball": []}


class DummyUglyStepScene(DummySceneBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hue = 0.0

    def list_properties(self):
        self.list_properties_calls += 1
        return {
            "Render": [
                NumericProperty(
                    self,
                    "hue",
                    title="Hue",
                    min=0.0,
                    max=1.0,
                    count=360,
                )
            ]
        }


class UnsupportedScene(DummySceneBase):
    @classmethod
    def supports_system(cls, system):
        return False


class DummyImageContainer:
    def __init__(self, size=(32, 24), scene_factory=None):
        self._size = size
        self._scene_factory = scene_factory
        self.scene = None
        self.uploaded_state = None
        self.saved_frames = 0
        self.attach_scene(scene_factory or (lambda _ctx: DummyScene()))

    @property
    def size(self):
        return self._size

    def attach_scene(self, scene_factory):
        self._scene_factory = scene_factory
        if callable(scene_factory):
            self.scene = scene_factory(None)
        else:
            self.scene = scene_factory
        return self

    def upload(self, state):
        self.uploaded_state = state
        if hasattr(self.scene, "upload"):
            self.scene.upload(state)
        return self

    def save(self):
        self.saved_frames += 1
        return Image.new(
            "RGB",
            self._size,
            color=((self.saved_frames * 23) % 255, 34, 56),
        )


def install_dummy_scene_registry(monkeypatch):
    registry = {
        "Layer": DummyScene,
        "General": DummyGeneralScene,
        "Ball": DummyBlochScene,
        "Unsupported": UnsupportedScene,
    }
    monkeypatch.setattr(
        camera_viewer_module,
        "_load_builtin_scene_registry",
        lambda: dict(registry),
    )
    return registry


def test_switch_scene_message_prefers_residency_manager():
    calls = []

    class DummyResidency:
        def switch_scene(self, scene_name):
            calls.append(scene_name)
            return True

    session = SimpleNamespace(
        _scene_residency_manager=DummyResidency(),
        switch_scene=lambda _scene_name: (_ for _ in ()).throw(
            AssertionError("session.switch_scene should not run")
        ),
    )

    outcome = camera_viewer_module._apply_bound_session_message(
        session,
        "switch_scene",
        {"scene_name": "Ball"},
    )

    assert outcome.handled is True
    assert outcome.publish_frame is True
    assert calls == ["Ball"]


def test_camera_commit_message_does_not_publish_properties():
    session = SimpleNamespace(
        commit_camera_interaction=lambda: True,
    )

    outcome = camera_viewer_module._apply_bound_session_message(
        session,
        "camera_commit",
        {},
    )

    assert outcome.handled is True
    assert outcome.publish_frame is True
    assert outcome.publish_camera is True
    assert outcome.publish_summary is True
    assert outcome.publish_properties is False


def test_apply_property_change_message_publishes_properties():
    calls = []

    session = SimpleNamespace(
        apply_property_input=lambda *args, **kwargs: calls.append((args, kwargs))
        or UpdateMask.REDRAW
    )

    outcome = camera_viewer_module._apply_bound_session_message(
        session,
        "apply_property_change",
        {
            "group_id": "Render",
            "prop_index": 2,
            "value": True,
            "expected_field": "flag",
            "expected_title": "Flag",
        },
    )

    assert outcome.handled is True
    assert outcome.publish_frame is True
    assert outcome.publish_camera is True
    assert outcome.publish_summary is True
    assert outcome.publish_properties is True
    assert calls == [
        (
            ("Render", 2, True),
            {
                "expected_field": "flag",
                "expected_title": "Flag",
            },
        )
    ]


def test_mgl_camera_is_importable_without_moderngl():
    import topovec.mgl as mgl

    assert mgl.Camera is Camera


def test_unit_camera_initializes_without_system_argument():
    camera = UnitCamera()

    assert np.allclose(camera.look_at, [0.0, 0.0, 0.0])
    assert np.allclose(camera.look_direction, [0.0, 0.0, -1.0])
    assert np.allclose(camera.look_up, [0.0, 1.0, 0.0])
    assert camera.look_scale == 1
    assert camera.perspective_is_on is True


def test_camera_state_roundtrip():
    camera = Camera.from_angles(
        theta=1.1,
        phi=0.7,
        alpha=0.2,
        scale=3.5,
        perspective=False,
        look_at=[1.0, 2.0, 3.0],
    )

    state = tvm.serialize_camera(camera)
    clone = Camera()
    tvm.restore_camera(clone, state)

    assert np.allclose(clone.look_at, camera.look_at)
    assert np.allclose(clone.look_direction, camera.look_direction)
    assert np.allclose(clone.look_up, camera.look_up)
    assert clone.look_scale == camera.look_scale
    assert clone.perspective_is_on is camera.perspective_is_on


def test_property_specs_expose_kind_field_and_invalidates():
    owner = SimpleNamespace(flag=True, value=0.25, mode="alpha")

    numeric = NumericProperty(
        owner,
        "value",
        title="Value",
        min=0.0,
        max=1.0,
        count=5,
        reset=True,
    )
    boolean = BooleanProperty(owner, "flag", title="Flag")
    listed = ListProperty(owner, "mode", title="Mode", values=["alpha", "beta"])

    assert numeric.spec() == {
        "kind": "numeric",
        "field": "value",
        "title": "Value",
        "value": 0.25,
        "invalidates": True,
        "param_key": None,
        "min": 0.0,
        "max": 1.0,
        "count": 5,
    }
    assert boolean.spec() == {
        "kind": "boolean",
        "field": "flag",
        "title": "Flag",
        "value": True,
        "invalidates": False,
        "param_key": None,
    }
    assert listed.spec() == {
        "kind": "list",
        "field": "mode",
        "title": "Mode",
        "value": "alpha",
        "invalidates": False,
        "param_key": None,
        "values": ["alpha", "beta"],
        "selected_index": 0,
    }


def test_numeric_property_rejects_count_one_for_non_degenerate_range():
    owner = SimpleNamespace(value=0.25)

    with pytest.raises(
        ValueError,
        match="count must be at least 2",
    ):
        NumericProperty(
            owner,
            "value",
            title="Value",
            min=0.0,
            max=1.0,
            count=1,
        )


def test_render_session_snapshot_contains_property_groups(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())

    snapshot = session.snapshot()

    assert snapshot["revision"] == 0
    assert snapshot["camera"]["can_undo"] is False
    assert snapshot["camera"]["can_redo"] is False
    assert "import topovec" not in snapshot["summary_text"]
    assert "ic = tv.mgl.render_prepare('Layer', system)" in snapshot["summary_text"]
    assert "ic.upload(field)" in snapshot["summary_text"]
    assert "ic.scene.set_camera(tv.mgl.Camera.from_angles(" in snapshot["summary_text"]
    assert snapshot["summary_text"].rstrip().endswith("ic.save()")
    assert snapshot["selected_scene"] == "Layer"
    assert snapshot["available_scenes"] == ["Layer", "General", "Ball"]

    group = snapshot["property_groups"][0]
    assert group["id"] == "Render"
    assert group["title"] == "Render"

    mode_prop = group["properties"][0]
    assert mode_prop["kind"] == "list"
    assert mode_prop["field"] == "mode"
    assert mode_prop["invalidates"] is True
    assert mode_prop["values"] == ["alpha", "beta"]

    level_prop = group["properties"][1]
    assert level_prop["kind"] == "numeric"
    assert level_prop["min"] == 0.0
    assert level_prop["max"] == 1.0
    assert level_prop["input_kind"] == "number"
    assert level_prop["row_kind"] == "inline"
    assert level_prop["default_value"] == 0.5
    assert level_prop["is_default"] is True
    assert level_prop["can_reset"] is True

    flag_prop = group["properties"][2]
    assert flag_prop["input_kind"] == "switch"
    assert flag_prop["row_kind"] == "inline"

    mode_prop = group["properties"][0]
    assert mode_prop["input_kind"] == "dropdown"
    assert mode_prop["row_kind"] == "inline"


def test_render_session_render_snippet_omits_default_properties(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())

    snippet = session.render_snippet()

    assert "ic.scene['Render']['Mode']" not in snippet
    assert "ic.scene['Render']['Level']" not in snippet
    assert "ic.scene['Render']['Flag']" not in snippet
    assert "ic.scene['Render']['Axis']" not in snippet
    assert "ic.scene['Render']['Follow view']" not in snippet


def test_render_session_render_snippet_includes_only_changed_properties(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())

    session.apply_property_change(
        "Render",
        2,
        True,
        expected_field="flag",
        expected_title="Flag",
    )
    snippet = session.render_snippet()

    assert "ic.scene['Render']['Flag'] = True" in snippet
    assert "ic.scene['Render']['Level']" not in snippet


def test_render_session_render_snippet_drops_line_when_property_returns_to_default(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())

    session.apply_property_change(
        "Render",
        2,
        True,
        expected_field="flag",
        expected_title="Flag",
    )
    session.apply_property_change(
        "Render",
        2,
        False,
        expected_field="flag",
        expected_title="Flag",
    )
    snippet = session.render_snippet()

    assert "ic.scene['Render']['Flag'] = True" not in snippet
    assert "ic.scene['Render']['Flag'] = False" not in snippet


def test_render_session_render_snippet_uses_compact_numeric_formatting(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())

    restore_camera(
        session.camera,
        serialize_camera(
            Camera.from_angles(
                theta=0.123456789,
                phi=1.23456789,
                alpha=-1.987654321,
                scale=12.3456789,
                perspective=False,
                look_at=[0.000123456789, -9876.54321, 0.0],
            )
        ),
    )

    snippet = session.render_snippet()

    assert "theta=0.123457" in snippet
    assert "phi=1.23457" in snippet
    assert "alpha=-1.98765" in snippet
    assert "scale=12.3457" in snippet
    assert "look_at=[0.000123457, -9876.54, 0]" in snippet


def test_render_session_summary_and_snippet_do_not_create_extra_scenes(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    create_calls = []
    original_create_scene = camera_viewer_module._viewer_session_core.ViewerSession._create_scene

    def counted_create_scene(self, scene_cls):
        create_calls.append(getattr(scene_cls, "__name__", repr(scene_cls)))
        return original_create_scene(self, scene_cls)

    monkeypatch.setattr(
        camera_viewer_module._viewer_session_core.ViewerSession,
        "_create_scene",
        counted_create_scene,
    )
    session = RenderSession(DummyImageContainer())

    snapshot = session.snapshot()
    snippet = session.render_snippet()

    assert "render_prepare('Layer', system)" in snapshot["summary_text"]
    assert "render_prepare('Layer', system)" in snippet
    assert create_calls == []


def test_render_session_camera_event_does_not_reread_property_groups(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())
    calls = session.scene.list_properties_calls

    changed = session.apply_camera_event(event_type="rotate", dx=0.25, dy=0.1)
    camera = session.camera_snapshot()
    properties = session.properties_snapshot()

    assert changed is True
    assert session.scene.list_properties_calls == calls
    assert camera["revision"] == 0
    assert properties["revision"] == 0
    assert camera["history_size"] == 1
    assert camera["has_pending_delta"] is True


def test_render_session_property_change_rereads_dynamic_metadata(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())
    calls = session.scene.list_properties_calls

    session.apply_property_change(
        "Render",
        0,
        "beta",
        expected_field="mode",
        expected_title="Mode",
    )
    snapshot = session.snapshot()
    level_prop = snapshot["property_groups"][0]["properties"][1]

    assert snapshot["revision"] == 1
    assert session.scene.list_properties_calls >= calls + 1
    assert level_prop["min"] == -2.0
    assert level_prop["max"] == 2.0
    assert snapshot["property_groups"][0]["properties"][0]["value"] == "beta"


def test_render_session_non_reset_property_change_patches_cache_locally(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())
    calls = session.scene.list_properties_calls

    session.apply_property_change(
        "Render",
        2,
        True,
        expected_field="flag",
        expected_title="Flag",
    )
    snapshot = session.properties_snapshot()

    assert session.scene.list_properties_calls == calls
    assert snapshot["property_groups"][0]["properties"][2]["value"] is True


def test_render_session_axis_switch_restores_cached_layer_camera(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())
    session.camera.set_basis(depth=[1, 1, 1], up=[0, 0, 1])
    before = session.snapshot()["camera"]

    session.apply_property_change(
        "Render",
        3,
        0.0,
        expected_field="axis",
        expected_title="Axis",
    )
    after = session.snapshot()["camera"]

    assert session.scene.camera_profile == "layer-x"
    assert camera_state_signature(before) != camera_state_signature(after)

    session.camera.set_basis(depth=[1, 1, 0], up=[0, 0, 1])
    layer_x_custom = session.snapshot()["camera"]
    session.apply_property_change(
        "Render",
        3,
        2.0,
        expected_field="axis",
        expected_title="Axis",
    )
    restored_z = session.snapshot()["camera"]
    assert session.scene.camera_profile == "layer-z"
    assert camera_state_signature(restored_z) == camera_state_signature(before)

    session.apply_property_change(
        "Render",
        3,
        0.0,
        expected_field="axis",
        expected_title="Axis",
    )
    restored_x = session.snapshot()["camera"]
    assert camera_state_signature(restored_x) == camera_state_signature(layer_x_custom)


def test_render_session_non_reset_property_camera_change_commits_current_profile(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())
    before = session.snapshot()["camera"]

    session.apply_property_change(
        "Render",
        4,
        True,
        expected_field="follow_view",
        expected_title="Follow view",
    )
    after = session.snapshot()["camera"]

    assert after["history_size"] == 1
    assert after["has_pending_delta"] is False
    assert camera_state_signature(before) != camera_state_signature(after)


def test_render_session_camera_commit_and_toggle_use_history_stubs(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())
    base = session.snapshot()["camera"]

    session.apply_camera_event(event_type="rotate", dx=0.25, dy=0.1)
    rotated = session.snapshot()["camera"]
    assert rotated["history_size"] == 1
    assert rotated["revision"] == 0
    assert rotated["has_pending_delta"] is True
    assert camera_state_signature(base) != camera_state_signature(rotated)

    session.commit_camera_interaction()
    committed = session.snapshot()["camera"]
    assert committed["history_size"] == 1
    assert committed["revision"] == 1
    assert committed["has_pending_delta"] is False
    assert camera_state_signature(committed) == camera_state_signature(rotated)

    session.apply_camera_button("toggle_perspective")
    toggled = session.snapshot()["camera"]
    assert toggled["history_size"] == 1
    assert toggled["can_undo"] is False
    assert toggled["perspective"] is False
    assert toggled["has_pending_delta"] is False

    assert camera_state_signature(base) != camera_state_signature(toggled)


def test_render_session_snap_button_snaps_camera(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())
    session.camera.set_basis(depth=[0.1, 0.0, 0.995], up=[1.0, 0.0, 0.0])

    changed = session.apply_camera_button("snap")
    snapped = session.snapshot()["camera"]

    assert changed is True
    assert snapped["revision"] == 1
    assert np.allclose(session.camera.look_direction, [0.0, 0.0, 1.0])
    assert snapped["has_pending_delta"] is False


def test_render_session_undo_redo_are_disabled(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())
    base = session.snapshot()["camera"]

    session.apply_camera_event(event_type="rotate", dx=0.25, dy=0.1)
    assert session.apply_camera_button("undo") is False
    assert session.apply_camera_button("redo") is False

    current = session.snapshot()["camera"]
    assert current["history_size"] == 1
    assert current["can_undo"] is False
    assert current["can_redo"] is False
    assert camera_state_signature(current) != camera_state_signature(base)


def test_render_session_switch_scene_reuses_state_and_restores_profile_camera(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    state = np.arange(6, dtype=np.float32).reshape((2, 3))
    session = RenderSession(
        DummyImageContainer(scene_factory=lambda _ctx: DummyGeneralScene()),
        state=state,
    )

    assert session.selected_scene == "General"
    general_initial_upload = session.scene.uploaded_state
    assert general_initial_upload is state

    session.camera.set_basis(depth=[1, 1, 1], up=[0, 0, 1])
    general_camera = session.snapshot()["camera"]

    assert session.switch_scene("Ball") is True
    assert session.selected_scene == "Ball"
    assert session.scene.uploaded_state is state
    ball_scene = session.scene

    session.camera.set_basis(depth=[0, 1, 1], up=[1, 0, 0])
    bloch_camera = session.snapshot()["camera"]

    assert session.switch_scene("General") is True
    restored_general = session.snapshot()["camera"]
    assert camera_state_signature(restored_general) == camera_state_signature(general_camera)

    assert session.switch_scene("Ball") is True
    assert session.scene is ball_scene
    restored_bloch = session.snapshot()["camera"]
    assert camera_state_signature(restored_bloch) == camera_state_signature(bloch_camera)


def test_render_session_render_snippet_updates_after_scene_switch(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())

    assert "render_prepare('Layer', system)" in session.render_snippet()

    session.switch_scene("Ball")
    snippet = session.render_snippet()

    assert "render_prepare('Ball', system)" in snippet
    assert "ic.scene['Render']" not in snippet


def test_render_session_shared_param_key_persists_across_scene_switch(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())

    session.apply_property_change(
        "Render",
        2,
        True,
        expected_field="flag",
        expected_title="Flag",
    )
    assert session.switch_scene("General") is True

    snapshot = session.properties_snapshot()
    general_props = snapshot["property_groups"][0]["properties"]

    assert general_props[1]["field"] == "flag"
    assert general_props[1]["value"] is True


def test_render_session_switch_scene_rejects_incompatible_builtin_scene(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    session = RenderSession(DummyImageContainer())

    with pytest.raises(ValueError, match="incompatible"):
        session.switch_scene("Unsupported")


def test_custom_transport_queue_drops_stale_camera_commit_before_later_delta():
    esm = _camera_viewport_widget_class()._esm

    assert "const dropPendingCameraCommits" in esm
    assert "dropPendingCameraCommits();" in esm
    assert re.search(
        r"const onPointerDown = .*?dropPendingCameraCommits\(\);",
        esm,
        re.DOTALL,
    )
    assert re.search(
        r"const onWheel = .*?dropPendingCameraCommits\(\);",
        esm,
        re.DOTALL,
    )
    assert 'queueEvent("camera_commit");' in esm


def test_viewport_widgets_offer_snap_action():
    viewport_esm = _camera_viewport_widget_class()._esm
    composite_esm = camera_viewer_module._composite_director_field_widget_class()._esm

    assert "Snap" in viewport_esm
    assert '"snap_camera"' in viewport_esm
    assert 'queueEvent("snap")' in viewport_esm
    assert "Snap" in composite_esm
    assert '"snap_camera"' in composite_esm


def test_marimo_widget_assets_are_available_as_package_resources():
    assets_module = importlib.import_module("topovec.marimo._assets")
    assets_root = importlib_resources.files("topovec.marimo.assets")

    viewport_js = assets_root.joinpath("camera_viewport.js").read_text(
        encoding="utf-8"
    )
    viewport_css = assets_root.joinpath("camera_viewport.css").read_text(
        encoding="utf-8"
    )
    inspect_js = assets_root.joinpath("director_field_inspect.js").read_text(
        encoding="utf-8"
    )
    inspect_css = assets_root.joinpath("director_field_inspect.css").read_text(
        encoding="utf-8"
    )

    assert "topovec-camera-viewer" in viewport_js
    assert ".topovec-camera-viewer" in viewport_css
    assert "topovec-composite-widget" in inspect_js
    assert ".topovec-composite-widget" in inspect_css
    assert viewport_js == assets_module.camera_viewport_esm()
    assert viewport_css == assets_module.camera_viewport_css()
    assert inspect_js == assets_module.director_field_inspect_esm()
    assert inspect_css == assets_module.director_field_inspect_css()


def test_anywidget_classes_load_assets_from_resource_loader():
    assets_module = importlib.import_module("topovec.marimo._assets")
    viewport_cls = _camera_viewport_widget_class()
    composite_cls = camera_viewer_module._composite_director_field_widget_class()

    assert viewport_cls._esm == assets_module.camera_viewport_esm()
    assert viewport_cls._css == assets_module.camera_viewport_css()
    assert composite_cls._esm == assets_module.director_field_inspect_esm()
    assert composite_cls._css == assets_module.director_field_inspect_css()


def test_composite_widget_uses_compact_panel_contract():
    widget_cls = camera_viewer_module._composite_director_field_widget_class()
    css = widget_cls._css
    esm = widget_cls._esm

    assert "grid-template-columns: minmax(0, 3fr) minmax(0, 1fr);" in css
    assert ".topovec-composite-rows" in css
    assert ".topovec-composite-row" in css
    assert ".topovec-composite-grid" not in css
    assert "buildCompactRow" in esm
    assert 'buildCompactRow("Reset"' not in esm
    assert 'buildCompactRow("Projection"' not in esm
    assert "other.open = false;" in esm
    assert 'summaryLabel.textContent = "Reproduce render";' in esm


def test_quantize_numeric_value_uses_count_resolution():
    prop_snapshot = {
        "min": 0.0,
        "max": 1.0,
        "count": 5,
        "value": 0.0,
    }

    assert _quantize_numeric_value(0.33, prop_snapshot) == 0.25
    assert _quantize_numeric_value(0.87, prop_snapshot) == 0.75


def test_quantize_numeric_value_falls_back_to_three_decimals():
    prop_snapshot = {
        "min": 1.0,
        "max": 0.5,
        "count": 2,
        "value": 0.75,
    }

    assert _effective_numeric_step(prop_snapshot) == 0.001
    assert _quantize_numeric_value(0.7334, prop_snapshot) == 0.733
    assert _quantize_numeric_value(0.7336, prop_snapshot) == 0.734


def test_inspect_view_prepares_container_from_state(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    import topovec.mgl as mgl

    captured = {}

    def fake_render_prepare(
        scenecls,
        system,
        camera_preset=2,
        scalex=False,
        imgsize=1024,
    ):
        captured["scenecls"] = scenecls
        captured["system"] = system
        captured["camera_preset"] = camera_preset
        captured["scalex"] = scalex
        captured["imgsize"] = imgsize
        return DummyImageContainer()

    monkeypatch.setattr(mgl, "render_prepare", fake_render_prepare)
    state = np.zeros((2, 3, 4, 3), dtype=np.float32)

    view = inspect_view(state)
    raw_widget = getattr(view.widget, "widget", view.widget)

    assert isinstance(view, DirectorFieldView)
    assert captured["scenecls"] == "Layer"
    assert tuple(int(v) for v in captured["system"].size) == (2, 3, 4)
    assert captured["camera_preset"] is None
    assert captured["scalex"] is False
    assert captured["imgsize"] == 512
    assert view.layout is view.widget
    assert hasattr(view.widget, "_mime_")
    assert raw_widget.properties_snapshot["selected_scene"] == "Layer"
    assert raw_widget.summary_rows == 8
    assert "render_prepare('Layer', system)" in raw_widget.summary_text
    assert view.session.selected_scene == "Layer"


def test_inspect_view_uses_explicit_system(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    import topovec.mgl as mgl

    captured = {}
    explicit_system = DummySystem(dim=3)

    def fake_render_prepare(
        scenecls,
        system,
        camera_preset=2,
        scalex=False,
        imgsize=1024,
    ):
        captured["scenecls"] = scenecls
        captured["system"] = system
        return DummyImageContainer()

    monkeypatch.setattr(mgl, "render_prepare", fake_render_prepare)

    view = inspect_view(
        np.zeros((2, 3, 4, 1, 3), dtype=np.float32),
        system=explicit_system,
        scene="Ball",
    )

    assert isinstance(view, DirectorFieldView)
    assert captured["scenecls"] == "Ball"
    assert captured["system"] is explicit_system
    assert view.session.selected_scene == "Ball"


def test_inspect_returns_layout(monkeypatch):
    sentinel = SimpleNamespace()
    view = SimpleNamespace(widget=sentinel, layout=sentinel)
    monkeypatch.setattr(
        camera_viewer_module,
        "inspect_view",
        lambda *args, **kwargs: view,
    )

    layout = inspect_layout(np.zeros((2, 3, 4, 3), dtype=np.float32))

    assert layout is sentinel
    assert layout._topovec_director_field_view is view


def test_inspect_widget_switch_scene_updates_properties_snapshot(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    import topovec.mgl as mgl

    monkeypatch.setattr(mgl, "render_prepare", lambda *args, **kwargs: DummyImageContainer())
    view = inspect_view(np.zeros((2, 3, 4, 3), dtype=np.float32))
    raw_widget = getattr(view.widget, "widget", view.widget)

    assert raw_widget.properties_snapshot["selected_scene"] == "Layer"

    raw_widget._handle_custom_message(
        None,
        {"type": "switch_scene", "scene_name": "Ball"},
        None,
    )

    assert view.session.selected_scene == "Ball"
    assert raw_widget.properties_snapshot["selected_scene"] == "Ball"
    assert raw_widget.properties_snapshot["property_groups"][0]["title"] == "Ball"
    assert raw_widget.camera_snapshot["revision"] == 1
    assert "render_prepare('Ball', system)" in raw_widget.summary_text


def test_render_session_scene_residency_eviction_preserves_gl_context():
    import topovec.mgl as mgl

    system = System.cubic((4, 4, 4))
    state = np.zeros((4, 4, 4, 3), dtype=np.float32)
    state[..., 2] = 1.0
    image_container = mgl.render_prepare(
        scenecls="Layer",
        system=system,
        imgsize=32,
    )
    manager = camera_viewer_module.SceneResidencyManager(
        image_container,
        scene_registry_loader=lambda: {
            "Layer": mgl.SCENES["Layer"],
            "Ball": mgl.SCENES["Ball"],
            "Level surface": mgl.SCENES["Level surface"],
        },
        session_factory=RenderSession,
        max_live_scenes=1,
    )

    try:
        session = manager.activate(
            system=system,
            state=state,
            preferred_scene_name="Layer",
        )

        assert manager.switch_scene("Ball") is True
        assert manager.switch_scene("Level surface") is True

        payload = session.frame_payload(image_encoder=lambda image: image.size)

        assert session.selected_scene == "Level surface"
        assert payload["image_data"] == image_container.size
        assert payload["projection_mode"] == "Perspective"
    finally:
        manager.release()


def test_inspect_widget_exposes_normalized_numeric_step(monkeypatch):
    import topovec.mgl as mgl

    monkeypatch.setattr(
        camera_viewer_module,
        "_load_builtin_scene_registry",
        lambda: {"Layer": DummyUglyStepScene},
    )
    monkeypatch.setattr(
        mgl,
        "render_prepare",
        lambda *args, **kwargs: DummyImageContainer(
            scene_factory=lambda _ctx: DummyUglyStepScene()
        ),
    )

    view = inspect_view(np.zeros((2, 3, 4, 3), dtype=np.float32))
    raw_widget = getattr(view.widget, "widget", view.widget)
    prop = raw_widget.properties_snapshot["property_groups"][0]["properties"][0]

    assert prop["title"] == "Hue"
    assert prop["step"] == pytest.approx(0.0028)


def test_inspect_widget_property_message_rebuilds_snapshot(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    import topovec.mgl as mgl

    monkeypatch.setattr(mgl, "render_prepare", lambda *args, **kwargs: DummyImageContainer())
    view = inspect_view(np.zeros((2, 3, 4, 3), dtype=np.float32))
    raw_widget = getattr(view.widget, "widget", view.widget)

    raw_widget._handle_custom_message(
        None,
        {
            "type": "apply_property_change",
            "group_id": "Render",
            "prop_index": 0,
            "value": "beta",
            "expected_field": "mode",
            "expected_title": "Mode",
        },
        None,
    )

    level_prop = raw_widget.properties_snapshot["property_groups"][0]["properties"][1]
    assert view.session.scene.get_mode() == "beta"
    assert level_prop["min"] == -2.0
    assert level_prop["max"] == 2.0
    assert "ic.scene['Render']['Mode'] = 'beta'" in raw_widget.summary_text


def test_inspect_widget_numeric_input_is_not_quantized(monkeypatch):
    install_dummy_scene_registry(monkeypatch)
    import topovec.mgl as mgl

    monkeypatch.setattr(mgl, "render_prepare", lambda *args, **kwargs: DummyImageContainer())
    view = inspect_view(np.zeros((2, 3, 4, 3), dtype=np.float32))
    raw_widget = getattr(view.widget, "widget", view.widget)

    raw_widget._handle_custom_message(
        None,
        {
            "type": "apply_property_change",
            "group_id": "Render",
            "prop_index": 1,
            "value": 0.26,
            "expected_field": "level",
            "expected_title": "Level",
        },
        None,
    )

    level_prop = raw_widget.properties_snapshot["property_groups"][0]["properties"][1]
    assert view.session.scene.level == 0.26
    assert level_prop["value"] == 0.26
    assert "ic.scene['Render']['Level'] = 0.26" in raw_widget.summary_text


def test_inspect_view_rejects_invalid_state_shape():
    with pytest.raises(ValueError, match="at least three spatial dimensions"):
        inspect_view(np.zeros((2, 3, 4), dtype=np.float32))

    with pytest.raises(ValueError, match="last state axis to have length 3"):
        inspect_view(np.zeros((2, 3, 4, 2), dtype=np.float32))


def test_marimo_namespace_exports_inspect_helpers():
    assert tvm.inspect is camera_viewer_module.inspect
    assert tvm.inspect_view is camera_viewer_module.inspect_view
    assert tvm.DirectorFieldView is camera_viewer_module.DirectorFieldView
    assert tvm.camera_viewer is camera_viewer_module.camera_viewer
    assert tvm.CameraViewportWidget is camera_viewer_module.CameraViewportWidget
    assert not hasattr(tvm, "CompositeDirectorFieldView")
    assert not hasattr(tvm, "render_director_field_view")
    assert not hasattr(tvm, "render_viewer")
    assert not hasattr(tvm, "render_layout")
    assert not hasattr(tvm, "sync_viewer_commands")


def test_camera_viewer_smoke():
    widget = tvm.camera_viewer(
        DummyImageContainer(),
        state=np.arange(6, dtype=np.float32).reshape((2, 3)),
    )
    raw_widget = getattr(widget, "widget", widget)

    assert hasattr(widget, "_mime_")
    assert raw_widget.use_custom_transport is False
    assert raw_widget.show_toolbar is True
    assert raw_widget.projection_mode == "Perspective"


def test_camera_viewport_widget_factory_returns_public_widget():
    widget = tvm.CameraViewportWidget(DummyImageContainer())

    assert widget.__class__.__name__ == "CameraViewportWidget"
    assert hasattr(widget, "set_frame_payload")
    assert widget.use_custom_transport is False


def test_pil_image_to_data_url_uses_png_data_uri():
    image = Image.new("RGB", (2, 2), color=(12, 34, 56))
    data_url = tvm.pil_image_to_data_url(image)

    assert data_url.startswith("data:image/png;base64,")
    assert len(data_url) > len("data:image/png;base64,")


def test_camera_summary_is_python_constructor_call():
    camera = Camera()
    summary = tvm.format_camera_summary(camera)

    assert summary == (
        "tv.mgl.Camera.from_angles("
        "theta=0, phi=0, alpha=-1.5708, scale=1, "
        "perspective=True, look_at=[0, 0, 0])"
    )
    assert "theta=0" in summary
    assert "phi=0" in summary
    assert "alpha=-1.5708" in summary
    assert "scale=1" in summary
    assert "perspective=True" in summary
    assert "look_at=[0, 0, 0]" in summary



def test_import():
    import topovec


def test_system_exposes_center_and_lengths():
    from topovec import System
    import numpy as np

    system = System.cubic(size=(2, 3, 4))

    assert np.allclose(system.center, [1.0, 1.5, 2.0])
    assert np.allclose(system.lengths, [2.0, 3.0, 4.0])


def test_plain_system_does_not_advertise_energy_mask_backend():
    from topovec import System
    from topovec.mgl.scene import VectorScene

    system = System.cubic(size=(2, 2, 2))

    assert VectorScene.supports_energy_mask_system(system) is False


def test_system_get_inner_triangles_uses_stored_triangulation_by_default():
    from topovec import System

    system = System.cubic(size=(2, 2, 2))
    triangles = system.get_inner_triangles()

    assert triangles.ndim == 2
    assert triangles.shape[1] == 3
    assert triangles.shape[0] > 0


def test_system_thinned_matches_field_shape_for_non_divisible_scalar_step():
    from topovec import System
    import numpy as np

    system = System.cubic(size=(121, 121, 21))
    data = np.zeros((121, 121, 21, 1, 3), dtype=np.float64)

    thin_data, thin_system = system.thinned(data=data, steps=4)

    assert thin_data.shape[:3] == (31, 31, 6)
    assert thin_data.shape[:3] == tuple(int(v) for v in thin_system.size)
    assert np.allclose(thin_system.primitives, system.primitives * 4)
    assert np.allclose(thin_system.representatives, system.representatives)


def test_system_thinned_matches_field_shape_for_vector_step_with_include():
    from topovec import System
    import numpy as np

    system = System.cubic(size=(5, 6, 7))
    data = np.zeros((5, 6, 7, 1, 3), dtype=np.float64)
    steps = np.array((4, 4, 4))
    include = np.array((3, 0, 1))

    thin_data, thin_system = system.thinned(data=data, steps=steps, include=include)

    assert thin_data.shape[:3] == (1, 2, 2)
    assert thin_data.shape[:3] == tuple(int(v) for v in thin_system.size)
    assert np.allclose(thin_system.primitives, system.primitives * steps[:, None])
    assert np.allclose(
        thin_system.representatives,
        system.representatives + include[None] @ system.primitives,
    )


def test_system_thinned_preserves_divisible_case_sizes():
    from topovec import System
    import numpy as np

    system = System.cubic(size=(120, 120, 20))
    data = np.zeros((120, 120, 20, 1, 3), dtype=np.float64)

    thin_data, thin_system = system.thinned(data=data, steps=4)

    assert thin_data.shape[:3] == (30, 30, 5)
    assert thin_data.shape[:3] == tuple(int(v) for v in thin_system.size)
    assert np.allclose(thin_system.primitives, system.primitives * 4)
    assert np.allclose(thin_system.representatives, system.representatives)



import moderngl as mgl
import numpy as np
from dataclasses import dataclass
from types import FunctionType

from .render import *
from .buffer import Float3DBuffer, IntBuffer, Field1DBuffer, Field3DBuffer
from ..core import NumericProperty, ListProperty, BooleanProperty


######################################################################################################


class LevelRender(StateRender):
    VERTEX = """
            #version 330

            in vec3 in_center; // center of atom
            in vec3 in_spin; // spin direction

            out VS_OUT {
                vec3 origin; // atom position in world coordiantes
                vec3 spin; // spin direction in spin space
            } vs_out;

            void main() {
                vec3 spin = normalize(in_spin);
                vs_out.origin = in_center;
                vs_out.spin = spin;
            } 
        """
    GEOMETRY = """
            #version 330 core

            const float M_PI = 3.1415926535897932384626433832795;


            layout (lines_adjacency) in;
            layout (triangle_strip, max_vertices = 16) out;

            uniform vec3 u_axis; // Zenith direction.
            uniform float u_level;
            uniform bool u_constant_theta;


            uniform mat4 u_view;
            uniform mat4 u_projection;
            uniform int u_subdivide=1;

            uniform vec3 u_camera; // camera position in world coordinates
            uniform float u_color_shift;
            uniform bool u_neglect_z;
            uniform bool u_inverted;

            in VS_OUT {
                vec3 origin;
                vec3 spin;
            } vs_in[];

            out vec3 v_position;
            out vec3 v_color;
            out vec3 v_light_direction;

            struct Input {
                vec3 origin;
                vec3 spin;
                float potential;
            };

            vec3 hsv2rgb(vec3 c) {
                vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
                vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
                return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
            }

            vec3 spin_to_color(vec3 spin) {
                float s = spin.z;
                float phi = (atan(spin.y, spin.x)/M_PI+1)/2;
                vec3 hsv = vec3( phi, clamp(1-0.9*s, 0, 1), clamp(1+0.6*s, 0, 1) );
                hsv.x += u_color_shift;                        
                return hsv2rgb(hsv);
            }

            void run0(Input a, Input b) {
                // f(t)=f(0)+t*(f(1)-f(0))=0
                // t = f(0)/(f(0)-f(1)))
                float t = a.potential/(a.potential-b.potential);
                if (!(0<t && t<1)) return;
                v_position = mix(a.origin, b.origin, t);
                vec3 spin = mix(a.spin, b.spin, t); 
                if (u_neglect_z) {
                    v_color = spin_to_color(vec3(spin.xy, 0));
                } else {
                    v_color = spin_to_color(spin);
                };

                v_light_direction = - normalize(mix(a.origin, b.origin, t)-u_camera);

                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
            }

            void run(Input a, Input b, Input c, Input d) {
                // There should be no point on the edge ab.
                run0(a, c);
                run0(b, c);
                run0(a, d);
                run0(b, d);
                run0(c, d);
            }

            void process(Input a0, Input a1, Input a2, Input a3) {
                     if (sign(a0.potential)==sign(a1.potential)) run(a0, a1, a2, a3);
                else if (sign(a0.potential)==sign(a2.potential)) run(a0, a2, a1, a3);
                else if (sign(a1.potential)==sign(a2.potential)) run(a1, a2, a3, a0);
                EndPrimitive();
            }

            void interpolate(in float t0, in float t1, in float t2, in float t3, out Input res) {
                res.origin = t0*vs_in[0].origin+t1*vs_in[1].origin+t2*vs_in[2].origin+t3*vs_in[3].origin;
                res.spin = normalize(t0*vs_in[0].spin+t1*vs_in[1].spin+t2*vs_in[2].spin+t3*vs_in[3].spin);
                if (u_constant_theta) {
                    // float theta = acos(dot(res.spin, u_axis));
                    // res.potential = 2*theta/M_PI - u_level - 1.0;

                    // res.potential = cos(theta-0.5*M_PI*u_level);
                
                    float theta = dot(res.spin, u_axis);
                    res.potential = theta - u_level;
                } else {
                    float phi = atan(res.spin.y, res.spin.x);
                    res.potential = cos(phi-M_PI*u_level);
                };
            }

            void main() {
                Input a0, a1, a2, a3;
                interpolate(1, 0, 0, 0, a0);
                interpolate(0, 1, 0, 0, a1);
                interpolate(0, 0, 1, 0, a2);
                interpolate(0, 0, 0, 1, a3);
                process(a0, a1, a2, a3);

                
                Input ac;
                interpolate(0.25, 0.25, 0.25, 0.25, ac);
                process(a0, a1, a2, ac);
                process(a1, a2, a3, ac);
                process(a2, a3, a0, ac);
                process(a3, a0, a1, ac);
                

                /*
                Input a01, a02, a03, a12, a13, a23;
                interpolate(0.5, 0.5, 0, 0, a01);
                interpolate(0.5, 0, 0.5, 0, a02);
                interpolate(0.5, 0, 0, 0.5, a03);
                interpolate(0, 0.5, 0.5, 0, a12);
                interpolate(0, 0.5, 0, 0.5, a13);
                interpolate(0, 0, 0.5, 0.5, a23);
                process(a0, a01, a02, a03);
                process(a1, a01, a12, a13);
                process(a2, a02, a12, a23);
                process(a3, a03, a13, a23);
                process(a01, a02, a12, ac);
                process(a01, a03, a13, ac);
                process(a02, a03, a23, ac);
                process(a12, a13, a23, ac);
                */
            }   
        """
    FRAGMENT = """
            #version 330

            uniform vec3 u_camera;
            uniform bool u_lighting;

            // in vec3 v_normal;
            in vec3 v_position;
            in vec3 v_color;
            in vec3 v_light_direction;

            out vec4 f_color;

            void main() {
                if (u_lighting) {
                    vec3 normal = normalize(cross(dFdx(v_position), dFdy(v_position)));
                    // vec3 normal = normalize(v_normal);
                    float inner_reflected = dot(reflect(normalize(v_position-u_camera), normal), v_light_direction);
                    float inner_disperse = dot(normal, v_light_direction);
                    float intensity = 0.4+0.6*clamp(inner_disperse, 0.0, 1.0) + 0.3*clamp(inner_reflected, 0.0, 1.0);
                    f_color = vec4(intensity*v_color.xyz, 1.0);
                    // f_color = vec4(abs(normal), 1.0); // DEBUG
                } else {
                     f_color = vec4(v_color.xyz, 1.0);
                };
            }

        """

    ANGLENAMES = ["polar angle", "azimuthal angle"]

    def __init__(
        self,
        coordinatesBuffer=None,
        maskBuffer=None,
        indexBuffer=None,
        scalarMaskBuffer=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        assert self._system.dim == 3

        self.u_view = self._prog["u_view"]
        self.u_projection = self._prog["u_projection"]
        self.u_color_shift = self._prog["u_color_shift"]
        self.u_level = self._prog["u_level"]
        self.u_constant_theta = self._prog["u_constant_theta"]
        self.u_lighting = self._prog["u_lighting"]
        self.u_axis = self._prog["u_axis"]
        self.u_neglect_z = self._prog["u_neglect_z"]

        self._coordinatesBuffer = coordinatesBuffer
        self._coordinatesBuffer.upload(self._system.spin_positions())

        self._idx = indexBuffer

        self.init_vao()

        self.lighting = True
        self.color_shift = 0.81
        self.level = 0.0
        self.constant_theta = True
        self.axis = (0, 0, 1)

    def init_vao(self):
        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self._state.vbo, "3f/v", "in_spin"),
                (self._coordinatesBuffer.vbo, "3f/v", "in_center"),
            ],
            self._idx.ibo,
        )

    @property
    def angle_name(self):
        return self.ANGLENAMES[0 if self.constant_theta else 1]

    @angle_name.setter
    def angle_name(self, name):
        self.constant_theta = name == self.ANGLENAMES[0]
        # self.init_vao()

    def render(
        self, scales=None, modelview=None, projection=None, camera_position=None
    ):
        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())
        self.u_camera.value = tuple(camera_position)

        self.u_axis.value = self.axis
        self.u_constant_theta.value = self.constant_theta
        self.u_neglect_z.value = self.constant_theta
        self.u_color_shift.value = self.color_shift
        self.u_level.value = self.level
        self.u_lighting.value = self.lighting

        self._ctx.enable_only(mgl.DEPTH_TEST)

        self._vao.render(mode=mgl.LINES_ADJACENCY, vertices=-1, first=0)

    def list_properties(self):
        return [
            BooleanProperty(self, "lighting", title="Lighting"),
            ListProperty(self, "angle_name", title="Constant", values=self.ANGLENAMES),
            # BooleanProperty(self, "constant_theta", title="Constant z-projection of spin"),
            NumericProperty(
                self, "level", title="Level", min=-1.0, max=1.0, count=1000
            ),
            NumericProperty(
                self, "color_shift", title="Hue", min=0.0, max=1.0, count=360
            ),
        ]


####################################################################################################################


@dataclass
class SurfaceType:
    name: str
    fn: FunctionType

class ScalarLevelRender(SystemRender):
    VERTEX='''
            #version 330

            in vec3 in_center; // center of atom
            in float in_value; // scalar field
            in vec3 in_color; // 
            
            // uniform float u_saturation = 0.5;
            // uniform float u_value = 1.0;

            out VS_OUT {
                vec3 origin; // atom position in world coordinates
                float value; // Scalar field value at the atom.
                vec3 color; 
            } vs_out;

            vec3 hsv2rgb(vec3 c) {
                vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
                vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
                return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
            }

            void main() {
                vs_out.origin = in_center;
                vs_out.value = in_value;
                vs_out.color = hsv2rgb(in_color);
                //vs_out.color = hsv2rgb(vec3(in_hue,u_saturation,u_value));
            } 
        '''
    GEOMETRY='''
            #version 330 core

            const float M_PI = 3.1415926535897932384626433832795;

            layout (lines_adjacency) in;
            layout (triangle_strip, max_vertices = 16) out;

            uniform vec3 u_min_pos;
            uniform vec3 u_max_pos;

            uniform float u_level;

            uniform mat4 u_view;
            uniform mat4 u_projection;
            uniform int u_subdivide=1;

            uniform vec3 u_camera; // camera position in world coordinates

            in VS_OUT {
                vec3 origin; // atom position in world coordinates
                float value; // Scalar field value at the atom.
                vec3 color; 
            } vs_in[];

            out vec3 v_position;
            out vec3 v_color;
            out vec3 v_light_direction;

            struct Input {
                vec3 origin;
                float potential;
                vec3 color; 
            };

            void run0(Input a, Input b) {
                // f(t)=f(0)+t*(f(1)-f(0))=0
                // t = f(0)/(f(0)-f(1)))
                float t = a.potential/(a.potential-b.potential);
                if (!(0<t && t<1)) return;
                v_position = mix(a.origin, b.origin, t);
                v_color = mix(a.color, b.color, t);
                v_light_direction = - normalize(mix(a.origin, b.origin, t)-u_camera);

                gl_Position = u_projection * u_view * vec4(v_position, 1.0);
                EmitVertex();
            }

            void run(Input a, Input b, Input c, Input d) {
                // There should be no point on the edge ab.
                run0(a, c);
                run0(b, c);
                run0(a, d);
                run0(b, d);
                run0(c, d);
            }

            void process(Input a0, Input a1, Input a2, Input a3) {
                     if (sign(a0.potential)==sign(a1.potential)) run(a0, a1, a2, a3);
                else if (sign(a0.potential)==sign(a2.potential)) run(a0, a2, a1, a3);
                else if (sign(a1.potential)==sign(a2.potential)) run(a1, a2, a3, a0);
                EndPrimitive();
            }

            void interpolate(in float t0, in float t1, in float t2, in float t3, out Input res) {
                res.origin = t0*vs_in[0].origin+t1*vs_in[1].origin+t2*vs_in[2].origin+t3*vs_in[3].origin;
                float potential = t0*vs_in[0].value+t1*vs_in[1].value+t2*vs_in[2].value+t3*vs_in[3].value;
                res.potential = potential - u_level;
                res.color = t0*vs_in[0].color+t1*vs_in[1].color+t2*vs_in[2].color+t3*vs_in[3].color;
            }

            // https://stackoverflow.com/questions/12751080/glsl-point-inside-box-test
            bool is_in_box(vec3 point, vec3 rectBottomLeft, vec3 rectTopRight) {
                vec3 bounded = clamp(point, rectBottomLeft, rectTopRight); 
                return point == bounded;
            }


            void main() {
                Input a0, a1, a2, a3;
                interpolate(1, 0, 0, 0, a0);
                interpolate(0, 1, 0, 0, a1);
                interpolate(0, 0, 1, 0, a2);
                interpolate(0, 0, 0, 1, a3);
                process(a0, a1, a2, a3);

                Input ac;
                interpolate(0.25, 0.25, 0.25, 0.25, ac);
                if(is_in_box(ac.origin, u_min_pos, u_max_pos)) {

                    process(a0, a1, a2, ac);
                    process(a1, a2, a3, ac);
                    process(a2, a3, a0, ac);
                    process(a3, a0, a1, ac);
                }
            }   
        '''
    FRAGMENT='''
            #version 330

            uniform vec3 u_min_pos;
            uniform vec3 u_max_pos;

            uniform vec3 u_camera;
            uniform bool u_lighting;

            // in vec3 v_normal;
            in vec3 v_position;
            in vec3 v_color;
            in vec3 v_light_direction;

            out vec4 f_color;

            // https://stackoverflow.com/questions/12751080/glsl-point-inside-box-test
            bool is_in_box(vec3 point, vec3 rectBottomLeft, vec3 rectTopRight) {
                vec3 bounded = clamp(point, rectBottomLeft, rectTopRight); 
                return point == bounded;
            }

            void main() {
                if (!is_in_box(v_position, u_min_pos, u_max_pos)) discard;
                if (u_lighting) {
                    vec3 normal = normalize(cross(dFdx(v_position), dFdy(v_position)));
                    // vec3 normal = normalize(v_normal);
                    float inner_reflected = dot(reflect(normalize(v_position-u_camera), normal), v_light_direction);
                    float inner_disperse = dot(normal, v_light_direction);
                    float intensity = 0.4+0.6*clamp(inner_disperse, 0.0, 1.0) + 0.3*clamp(inner_reflected, 0.0, 1.0);
                    f_color = vec4(intensity*v_color.xyz, 1.0);
                    // f_color = vec4(abs(normal), 1.0); // DEBUG
                } else {
                     f_color = vec4(v_color.xyz, 1.0);
                };
            }

        '''


    def __init__(self, coordinatesBuffer=None, maskBuffer=None, indexBuffer=None, scalarMaskBuffer=None, **kwargs):
        super().__init__(**kwargs)

        self.SURFACES_TYPES = [
            SurfaceType(name='Umbilique line', fn=self.make_delta), 
        ]

        assert self._system.dim==3

        self.u_min_pos = self._prog['u_min_pos']
        self.u_max_pos = self._prog['u_max_pos']

        self.u_view = self._prog['u_view']
        self.u_projection = self._prog['u_projection']
        self.u_level = self._prog['u_level']
        self.u_lighting = self._prog['u_lighting']

        self._coordinatesBuffer = coordinatesBuffer
        coords = self._system.spin_positions()
        self._min_pos = np.min(coords, axis=(0,1,2,3))
        self._max_pos = np.max(coords, axis=(0,1,2,3))

        self._coordinatesBuffer.upload(coords)

        self._potentialBuffer = Field1DBuffer(mglctx=self._ctx, system=self._system)
        self._colorBuffer = Field3DBuffer(mglctx=self._ctx, system=self._system)

        self._idx = indexBuffer

        self.init_vao()

        self._np_state = None

        self.lighting = True
        self.level = -2.
        self.mode = 0 
        self.border = 0.05

        self._max_potential = 0
        self._min_potential = -8

        

    def init_vao(self):
        self._vao = self._ctx.vertex_array(self._prog, 
            [
                (self._potentialBuffer.vbo, 'f/v', 'in_value'),
                (self._coordinatesBuffer.vbo, '3f/v', 'in_center'),
                (self._colorBuffer.vbo, '3f/v', 'in_color'),
            ],
                self._idx.ibo
            )

    @property
    def surface_name(self):
        return self.SURFACES_TYPES[self.mode].name

    @surface_name.setter
    def surface_name(self, name):
        idx = list(n for n, surface in enumerate(self.SURFACES_TYPES) if surface.name==name)
        assert len(idx)==1, f"Surface type {name} correspond to multiple indices {idx}"
        self.mode = idx[0]
        # Update 
        self.recompute()
        # self.init_vao()


    def render(self, scales=None, modelview=None, projection=None, camera_position=None):
        self.u_view.value = tuple( modelview.T.flatten() )
        self.u_projection.value = tuple( projection.T.flatten() )
        self.u_camera.value = tuple(camera_position)

        delta_pos = self._max_pos-self._min_pos
        self.u_min_pos.value = tuple(self._min_pos+delta_pos*self.border)
        self.u_max_pos.value = tuple(self._max_pos-delta_pos*self.border)
        self.u_level.value = self.level 
        self.u_lighting.value = self.lighting 

        self._ctx.enable_only(mgl.DEPTH_TEST)

        self._vao.render(mode=mgl.LINES_ADJACENCY, vertices=-1, first=0)

    def list_properties(self):
        return [
            BooleanProperty(self, "lighting", title="Lighting"),
            ListProperty(self, "surface_name", title="Mode", values=list(m.name for m in self.SURFACES_TYPES) ),
            NumericProperty(self, "level", title="Level", min=self._min_potential, max=self._max_potential, count=1000),
            NumericProperty(self, "border", title="Border", min=0.0, max=0.5, count=1000),
        ]

    def upload(self, nn):
        assert isinstance(nn, np.ndarray)
        nn = nn.astype(np.float64)
        self._np_state = nn/np.sqrt(np.sum(nn**2,axis=-1,keepdims=True)+1e-32)
        self.recompute()

    def recompute(self):
        pot, color = self.SURFACES_TYPES[self.mode].fn()
        self._potentialBuffer.upload( pot )
        self._colorBuffer.upload( color )

    def make_delta(self):
        nn = self._np_state
        dn = np.stack(list(np.roll(nn, shift=1, axis=a)-np.roll(nn, shift=-1, axis=a) for a in range(3)), axis=-1)

        P = np.eye(3) - np.einsum('...i,...j->...ij', nn, nn)
        tmp = np.einsum('...ij,...jk->...ik',    P,    dn)
        N   = np.einsum('...ij,...jk->...ik', tmp,   P)
        X = 0.5*(N + np.swapaxes(N, -2, -1))

        T1 = np.einsum('...ii->...', X)               # = Tr(N)
        T2 = np.einsum('...ij,...ji->...', X, X)      # = Tr(N^2)
        potential = (2*T2 - T1*T1)
        potential = np.log10(potential)
        self._max_potential = np.max(potential)
        self._min_potential = np.min(potential[~np.isinf(potential)])

        a = X[...,0,0]-X[...,1,1] 
        b = X[...,1,0]+X[...,0,1]

        hue = np.arctan2(b,a)/(2*np.pi)
        saturation = np.full_like(hue, 1.)
        # saturation = 1-(T1*T1)/(8*mx)

        color = np.stack([hue,saturation,np.full_like(hue, 1.)],axis=-1)
        return potential.astype(np.float32), color.astype(np.float32)

####################################################################################################################




"""Shared live-scene residency management for desktop and notebook viewers."""

from __future__ import annotations

import logging
from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Any, Callable

import numpy as np

from .session import (
    ViewerSession,
    ViewerSessionResidencySnapshot,
    adapt_viewer_host,
)

logger = logging.getLogger(__name__)


def _scene_supports_system(scene_cls: Any, system: Any) -> bool:
    supports = getattr(scene_cls, "supports_system", None)
    if not callable(supports):
        return True
    return bool(supports(system))


def _system_attribute_key(value: Any) -> Any:
    if value is None:
        return None
    array = np.asarray(value)
    if array.ndim == 0:
        return array.item()
    return tuple(int(v) for v in array.shape), tuple(array.reshape(-1).tolist())


def system_cache_key(system: Any) -> tuple[Any, ...] | None:
    """Compute the structural cache key shared by Qt and marimo viewers."""
    if system is None:
        return None
    return (
        type(system),
        *(
            _system_attribute_key(getattr(system, name, None))
            for name in ("size", "primitives", "representatives", "dim", "rank")
        ),
    )


@dataclass(slots=True)
class _SystemResidencyRecord:
    system: Any
    snapshot: ViewerSessionResidencySnapshot | None = None
    live_scenes: dict[str, Any] = field(default_factory=dict)


class SceneResidencyManager:
    """Manage per-system snapshots and a global LRU of live scenes."""

    def __init__(
        self,
        host: Any,
        *,
        scene_registry_loader: Callable[[], dict[str, Any]],
        session_factory: Callable[..., ViewerSession] = ViewerSession,
        session_kwargs: dict[str, Any] | None = None,
        max_live_scenes: int = 2,
    ):
        if int(max_live_scenes) < 1:
            raise ValueError("max_live_scenes must be at least 1.")
        self._host = adapt_viewer_host(host)
        self._scene_registry_loader = scene_registry_loader
        self._session_factory = session_factory
        self._session_kwargs = {} if session_kwargs is None else dict(session_kwargs)
        self._max_live_scenes = int(max_live_scenes)
        self._records: dict[tuple[Any, ...], _SystemResidencyRecord] = {}
        self._active_system_key: tuple[Any, ...] | None = None
        self._active_session: ViewerSession | None = None
        self._live_scene_lru: OrderedDict[tuple[tuple[Any, ...], str], None] = (
            OrderedDict()
        )
        self._last_activation_mode = "new_scene"

    @property
    def max_live_scenes(self) -> int:
        return self._max_live_scenes

    @property
    def session(self) -> ViewerSession | None:
        return self._active_session

    @property
    def last_activation_mode(self) -> str:
        return self._last_activation_mode

    def _scene_registry(self) -> dict[str, Any]:
        return dict(self._scene_registry_loader())

    def _resolve_scene_name(
        self,
        *,
        system: Any,
        record: _SystemResidencyRecord,
        preferred_scene_name: str | None,
    ) -> str:
        registry = self._scene_registry()
        supported = [
            str(name)
            for name, scene_cls in registry.items()
            if _scene_supports_system(scene_cls, system)
        ]
        if not supported:
            raise RuntimeError("No built-in scenes are compatible with this system.")

        snapshot_scene = (
            None if record.snapshot is None else record.snapshot.selected_scene
        )
        for candidate in (snapshot_scene, preferred_scene_name):
            if candidate and candidate in supported:
                return str(candidate)
        return supported[0]

    def _actual_live_scene_keys(self) -> set[tuple[tuple[Any, ...], str]]:
        keys: set[tuple[tuple[Any, ...], str]] = set()
        for system_key, record in self._records.items():
            for scene_name in record.live_scenes:
                keys.add((system_key, str(scene_name)))
        if self._active_session is not None and self._active_system_key is not None:
            for scene_name in self._active_session.live_scene_names():
                keys.add((self._active_system_key, str(scene_name)))
        return keys

    def _active_scene_key(self) -> tuple[tuple[Any, ...], str] | None:
        if self._active_session is None or self._active_system_key is None:
            return None
        scene_name = self._active_session.selected_scene
        if scene_name is None:
            return None
        return self._active_system_key, str(scene_name)

    def _reconcile_lru(self) -> None:
        actual = self._actual_live_scene_keys()
        for key in list(self._live_scene_lru.keys()):
            if key not in actual:
                self._live_scene_lru.pop(key, None)
        for key in actual:
            self._live_scene_lru.setdefault(key, None)

    def _mark_scene_used(self, system_key: tuple[Any, ...], scene_name: str) -> None:
        scene_key = (system_key, str(scene_name))
        self._reconcile_lru()
        self._live_scene_lru.setdefault(scene_key, None)
        self._live_scene_lru.move_to_end(scene_key)

    def _release_scene(self, scene: Any) -> None:
        release = getattr(scene, "release", None)
        if callable(release):
            release()

    def _enforce_live_scene_limit(self) -> None:
        self._reconcile_lru()
        pinned_key = self._active_scene_key()
        while len(self._live_scene_lru) > self._max_live_scenes:
            evicted = False
            for scene_key in list(self._live_scene_lru.keys()):
                if scene_key == pinned_key:
                    continue
                system_key, scene_name = scene_key
                if system_key == self._active_system_key and self._active_session is not None:
                    if self._active_session.evict_live_scene(scene_name):
                        logger.debug(
                            "Evicted live scene from active session system_key=%s scene=%s",
                            system_key,
                            scene_name,
                        )
                        self._live_scene_lru.pop(scene_key, None)
                        evicted = True
                        break
                    self._live_scene_lru.pop(scene_key, None)
                    continue
                record = self._records.get(system_key)
                if record is None:
                    self._live_scene_lru.pop(scene_key, None)
                    continue
                scene = record.live_scenes.pop(scene_name, None)
                if scene is None:
                    self._live_scene_lru.pop(scene_key, None)
                    continue
                logger.debug(
                    "Evicted live scene from inactive record system_key=%s scene=%s",
                    system_key,
                    scene_name,
                )
                self._release_scene(scene)
                self._live_scene_lru.pop(scene_key, None)
                evicted = True
                break
            if not evicted:
                break
        self._reconcile_lru()

    def _park_session(
        self,
        system_key: tuple[Any, ...],
        session: ViewerSession,
    ) -> None:
        record = self._records.setdefault(
            system_key,
            _SystemResidencyRecord(system=getattr(session.scene, "system", None)),
        )
        record.snapshot = session.export_residency_snapshot()
        record.live_scenes = session.export_live_scenes()
        logger.debug(
            "Parked viewer session system_key=%s selected_scene=%s live_scenes=%s",
            system_key,
            record.snapshot.selected_scene if record.snapshot is not None else None,
            sorted(record.live_scenes),
        )
        self._reconcile_lru()

    def activate(
        self,
        *,
        system: Any,
        state: np.ndarray | None,
        preferred_scene_name: str | None = None,
    ) -> ViewerSession:
        target_key = system_cache_key(system)
        if target_key is None:
            raise RuntimeError("Scene residency requires a non-null system.")

        if self._active_session is not None and self._active_system_key == target_key:
            self._active_session.set_render_state(state)
            active_scene = self._active_session.selected_scene
            if active_scene is not None:
                self._mark_scene_used(target_key, active_scene)
            self._last_activation_mode = "active_session"
            self._enforce_live_scene_limit()
            return self._active_session

        record = self._records.setdefault(
            target_key,
            _SystemResidencyRecord(system=system),
        )
        resolved_system = record.system
        scene_name = self._resolve_scene_name(
            system=resolved_system,
            record=record,
            preferred_scene_name=preferred_scene_name,
        )
        registry = self._scene_registry()
        scene = record.live_scenes.get(scene_name)
        activation_mode = "live_scene" if scene is not None else "new_scene"
        if scene is None:
            scene_cls = registry[scene_name]
            scene = self._host.create_scene(scene_cls, resolved_system)

        previous_key = self._active_system_key
        previous_session = self._active_session

        try:
            self._host.attach_scene(scene)
            session = self._session_factory(
                self._host,
                state=state,
                **self._session_kwargs,
            )
            session.seed_live_scenes(record.live_scenes)
            if record.snapshot is not None:
                session.restore_residency_snapshot(record.snapshot)
        except Exception:
            if previous_session is not None:
                self._host.attach_scene(previous_session.scene)
            raise

        if previous_key is not None and previous_session is not None:
            self._park_session(previous_key, previous_session)

        record.live_scenes = {}
        self._active_system_key = target_key
        self._active_session = session
        setattr(session, "_scene_residency_manager", self)
        self._last_activation_mode = activation_mode
        self._mark_scene_used(target_key, session.selected_scene or scene_name)
        self._enforce_live_scene_limit()
        logger.debug(
            "Activated residency-managed session mode=%s system_key=%s selected_scene=%s",
            activation_mode,
            target_key,
            session.selected_scene,
        )
        return session

    def switch_scene(self, scene_name: str) -> bool:
        if self._active_session is None or self._active_system_key is None:
            raise RuntimeError("No active viewer session to switch.")
        changed = self._active_session.switch_scene(scene_name)
        if not changed:
            return False
        self._mark_scene_used(self._active_system_key, scene_name)
        self._enforce_live_scene_limit()
        return True

    def deactivate(self) -> None:
        if self._active_session is None or self._active_system_key is None:
            return
        self._park_session(self._active_system_key, self._active_session)
        self._active_session = None
        self._active_system_key = None
        self._reconcile_lru()

    def release(self) -> None:
        if self._active_session is not None:
            self._active_session.release_live_scenes()
            self._active_session = None
            self._active_system_key = None
        for record in self._records.values():
            for scene in record.live_scenes.values():
                self._release_scene(scene)
            record.live_scenes = {}
            record.snapshot = None
        self._records.clear()
        self._live_scene_lru.clear()


__all__ = ["SceneResidencyManager", "system_cache_key"]



"""Qt OpenGL viewport and live ModernGL host."""

from __future__ import annotations

from contextlib import contextmanager
import logging
from typing import Any

import moderngl as mgl
from PIL import Image
from PySide6 import QtCore, QtGui
from PySide6.QtOpenGLWidgets import QOpenGLWidget

from ...mgl.target import OwnedRenderTarget, RenderTarget
from .theme import ThemeSpec, theme_spec_for_mode

logger = logging.getLogger(__name__)


class QtOpenGLLoader:
    """ModernGL loader backed by the current Qt OpenGL context."""

    def __init__(self, context: QtGui.QOpenGLContext):
        if context is None:
            raise RuntimeError("Qt OpenGL context is not available.")
        self._context = context

    def load_opengl_function(self, name: str) -> int:
        proc = self._context.getProcAddress(str(name))
        if proc is None:
            return 0
        try:
            return int(proc)
        except TypeError:
            return 0

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return None

    def release(self):
        return None


class QtViewportHost:
    """Live viewer host backed by a :class:`QOpenGLWidget`."""

    def __init__(self, viewport: "QtViewport"):
        self._viewport = viewport

    @property
    def scene(self) -> Any:
        return self._viewport._scene

    @property
    def size(self) -> tuple[int, int]:
        return self._viewport.current_render_size()

    def _context_description(self) -> str:
        current = QtGui.QOpenGLContext.currentContext()
        widget = self._viewport.context()
        return (
            f"current_context_present={current is not None} "
            f"widget_context_present={widget is not None} "
            f"current_matches_widget={bool(current is not None and widget is not None and current == widget)}"
        )

    def _make_current(self, reason: str) -> None:
        logger.debug("makeCurrent begin reason=%s %s", reason, self._context_description())
        self._viewport.makeCurrent()
        logger.debug("makeCurrent end reason=%s %s", reason, self._context_description())

    def _done_current(self, reason: str) -> None:
        logger.debug("doneCurrent begin reason=%s %s", reason, self._context_description())
        self._viewport.doneCurrent()
        logger.debug("doneCurrent end reason=%s %s", reason, self._context_description())

    def create_scene(self, scene_cls: Any, system: Any) -> Any:
        if self._viewport._ctx is None:
            raise RuntimeError("OpenGL context is not initialized yet.")
        logger.debug(
            "Creating scene scene_cls=%s system_size=%s %s",
            getattr(scene_cls, "__name__", repr(scene_cls)),
            tuple(int(v) for v in getattr(system, "size", ())),
            self._context_description(),
        )
        self._make_current("create_scene")
        try:
            scene = scene_cls(mglctx=self._viewport._ctx, system=system)
        except Exception:
            logger.exception(
                "Failed to create scene scene_cls=%s",
                getattr(scene_cls, "__name__", repr(scene_cls)),
            )
            raise
        finally:
            self._done_current("create_scene")
        logger.debug(
            "Created scene scene_cls=%s scene_type=%s",
            getattr(scene_cls, "__name__", repr(scene_cls)),
            type(scene).__name__,
        )
        return scene

    def attach_scene(self, scene: Any) -> Any:
        logger.debug(
            "Attaching scene scene_type=%s %s",
            type(scene).__name__,
            self._context_description(),
        )
        self._viewport._scene = scene
        return self

    def upload(self, state) -> Any:
        if self._viewport._scene is None:
            raise RuntimeError("No scene is attached.")
        logger.debug(
            "Uploading state scene_type=%s shape=%s dtype=%s %s",
            type(self._viewport._scene).__name__,
            tuple(int(v) for v in getattr(state, "shape", ())),
            getattr(state, "dtype", None),
            self._context_description(),
        )
        self._make_current("upload")
        try:
            self._viewport._scene.upload(state)
        except Exception:
            logger.exception(
                "Failed to upload state into scene scene_type=%s",
                type(self._viewport._scene).__name__,
            )
            raise
        finally:
            self._done_current("upload")
        return self

    @contextmanager
    def scene_mutation_scope(self, reason: str):
        self._make_current(reason)
        try:
            yield self
        finally:
            self._done_current(reason)

    def render_image(
        self,
        *,
        size: tuple[int, int] | None = None,
    ) -> Image.Image:
        if self._viewport._ctx is None or self._viewport._scene is None:
            raise RuntimeError("Viewport is not ready to render.")

        requested = (
            self._viewport.current_render_size()
            if size is None
            else (int(size[0]), int(size[1]))
        )
        logger.debug(
            "Rendering image size=%s scene_type=%s %s",
            requested,
            type(self._viewport._scene).__name__,
            self._context_description(),
        )
        self._make_current("render_image")
        target = OwnedRenderTarget(self._viewport._ctx, requested)
        try:
            self._viewport._prepare_render_target("render_image", target=target)
            self._viewport._scene.render_to(target)
            self._viewport._ctx.finish()
            data = target.read_rgb()
        finally:
            target.release()
            self._done_current("render_image")
        image = Image.frombytes("RGB", requested, data)
        return image.transpose(Image.FLIP_TOP_BOTTOM)


class QtViewport(QOpenGLWidget):
    """Central live viewport for the desktop viewer."""

    host_ready = QtCore.Signal()
    session_state_changed = QtCore.Signal(bool)
    runtime_error = QtCore.Signal(str)

    def __init__(self, parent: QtCore.QObject | None = None):
        super().__init__(parent)
        self._ctx: mgl.Context | None = None
        self._scene = None
        self._session = None
        self._mouse_pointer: tuple[float, float] | None = None
        self._drag_changed = False
        self._paint_error_signature: tuple[str, str] | None = None
        app = QtGui.QGuiApplication.instance()
        self._theme_spec = theme_spec_for_mode("system", app) if app is not None else None
        self.host = QtViewportHost(self)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_OpaquePaintEvent, True)
        self.setAutoFillBackground(False)
        self.setMouseTracking(True)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setUpdateBehavior(QOpenGLWidget.UpdateBehavior.NoPartialUpdate)

    @staticmethod
    def create_surface_format() -> QtGui.QSurfaceFormat:
        fmt = QtGui.QSurfaceFormat()
        fmt.setRenderableType(QtGui.QSurfaceFormat.RenderableType.OpenGL)
        fmt.setVersion(3, 3)
        fmt.setProfile(QtGui.QSurfaceFormat.OpenGLContextProfile.CoreProfile)
        fmt.setSwapInterval(1)
        fmt.setDepthBufferSize(24)
        fmt.setSamples(4)
        return fmt

    def current_render_size(self) -> tuple[int, int]:
        dpr = max(float(self.devicePixelRatioF()), 1.0)
        return (
            max(1, int(round(self.width() * dpr))),
            max(1, int(round(self.height() * dpr))),
        )

    @staticmethod
    def _enum_name(enum_value) -> str:
        try:
            return str(enum_value).rsplit(".", 1)[-1]
        except Exception:
            return repr(enum_value)

    @classmethod
    def _surface_format_snapshot(cls, fmt: QtGui.QSurfaceFormat | None) -> dict[str, object] | None:
        if fmt is None:
            return None
        return {
            "version": (int(fmt.majorVersion()), int(fmt.minorVersion())),
            "profile": cls._enum_name(fmt.profile()),
            "renderable_type": cls._enum_name(fmt.renderableType()),
            "swap_interval": int(fmt.swapInterval()),
            "samples": int(fmt.samples()),
            "depth_buffer_size": int(fmt.depthBufferSize()),
            "stencil_buffer_size": int(fmt.stencilBufferSize()),
            "alpha_buffer_size": int(fmt.alphaBufferSize()),
        }

    def _runtime_diagnostics(self, qt_context: QtGui.QOpenGLContext | None) -> dict[str, object]:
        info = dict(getattr(self._ctx, "info", {}) or {})
        actual_format = None
        is_opengles = None
        if qt_context is not None:
            format_getter = getattr(qt_context, "format", None)
            if callable(format_getter):
                actual_format = format_getter()
            gles_getter = getattr(qt_context, "isOpenGLES", None)
            if callable(gles_getter):
                try:
                    is_opengles = bool(gles_getter())
                except Exception:
                    is_opengles = None
        return {
            "vendor": info.get("GL_VENDOR"),
            "renderer": info.get("GL_RENDERER"),
            "version": info.get("GL_VERSION"),
            "glsl": info.get("GL_SHADING_LANGUAGE_VERSION"),
            "version_code": getattr(self._ctx, "version_code", None),
            "default_framebuffer": int(self.defaultFramebufferObject() or 0),
            "is_opengles": is_opengles,
            "requested_format": self._surface_format_snapshot(self.format()),
            "actual_format": self._surface_format_snapshot(actual_format),
        }

    @property
    def is_ready(self) -> bool:
        return self._ctx is not None

    def _current_widget_target(self) -> RenderTarget:
        if self._ctx is None:
            raise RuntimeError("OpenGL context is not initialized yet.")
        framebuffer_id = int(self.defaultFramebufferObject())
        if framebuffer_id <= 0:
            raise RuntimeError("QOpenGLWidget framebuffer is not initialized yet.")
        framebuffer = self._ctx.detect_framebuffer(framebuffer_id)
        framebuffer_size = tuple(
            int(v) for v in getattr(framebuffer, "size", self.current_render_size())
        )
        return RenderTarget(ctx=self._ctx, fbo=framebuffer, size=framebuffer_size)

    def _prepare_render_target(self, reason: str, *, target: RenderTarget) -> None:
        if self._ctx is None:
            raise RuntimeError("OpenGL context framebuffer is not initialized yet.")
        target.bind()
        blend_func = mgl.DEFAULT_BLENDING
        blend_equation = mgl.FUNC_ADD
        self._ctx.scissor = None
        target.fbo.scissor = None
        target.fbo.viewport = (0, 0, *target.size)
        target.fbo.color_mask = (True, True, True, True)
        target.fbo.depth_mask = True
        self._ctx.enable_only(mgl.NOTHING)
        self._ctx.blend_func = blend_func
        self._ctx.blend_equation = blend_equation
        self._ctx.viewport = (0, 0, *target.size)
        logger.debug(
            "Prepared render target reason=%s framebuffer=%s render_size=%s ctx_scissor=%s fbo_scissor=%s fbo_viewport=%s color_mask=%s depth_mask=%s blend_func=%s blend_equation=%s",
            reason,
            getattr(target.fbo, "glo", None),
            target.size,
            getattr(self._ctx, "scissor", None),
            getattr(target.fbo, "scissor", None),
            getattr(target.fbo, "viewport", None),
            getattr(target.fbo, "color_mask", None),
            getattr(target.fbo, "depth_mask", None),
            blend_func,
            blend_equation,
        )

    def set_session(self, session) -> None:
        self._session = session

    def apply_theme(self, spec: ThemeSpec | None) -> None:
        self._theme_spec = spec
        self.update()

    def _empty_clear_color(self) -> tuple[float, float, float, float]:
        if self._theme_spec is not None:
            return self._theme_spec.viewport_clear_rgba
        return (1.0, 1.0, 1.0, 1.0)

    def clear_session(self) -> None:
        self._session = None
        self._scene = None
        self.update()

    def shutdown(self) -> None:
        if self._ctx is None:
            self._session = None
            self._scene = None
            return
        logger.info("Shutting down topovec view OpenGL viewport")
        try:
            self.makeCurrent()
            self._session = None
            self._scene = None
            try:
                self._ctx.gc()
            except Exception:
                logger.debug("ModernGL context gc failed during shutdown.", exc_info=True)
            self._ctx.release()
        except Exception:
            logger.exception("Failed while shutting down topovec view OpenGL viewport")
        finally:
            self._ctx = None
            try:
                self.doneCurrent()
            except Exception:
                logger.debug("doneCurrent failed during viewport shutdown.", exc_info=True)

    def initializeGL(self) -> None:
        qt_context = QtGui.QOpenGLContext.currentContext() or self.context()
        logger.debug(
            "initializeGL begin current_context_present=%s widget_context_present=%s render_size=%s dpr=%.3f",
            QtGui.QOpenGLContext.currentContext() is not None,
            self.context() is not None,
            self.current_render_size(),
            max(float(self.devicePixelRatioF()), 1.0),
        )
        try:
            loader = QtOpenGLLoader(qt_context)
            logger.debug("Initializing ModernGL from Qt OpenGL context via getProcAddress()")
            mgl.init_context(loader=loader)
            self._ctx = mgl.get_context()
            logger.debug("ModernGL context initialized through Qt loader")
        except Exception:
            logger.exception("Qt-based ModernGL initialization failed; falling back to default loader")
            try:
                self._ctx = mgl.create_context()
            except Exception:
                logger.exception("Failed to initialize ModernGL context for topovec view viewport")
                self.runtime_error.emit(
                    "Failed to initialize the OpenGL viewport. "
                    "Qt OpenGL context initialization failed and moderngl fallback "
                    "could not load system OpenGL libraries. See Log for details."
                )
                self._ctx = None
                return
        if self._ctx is None:
            self.runtime_error.emit("Failed to initialize the OpenGL viewport. See Log for details.")
            return
        diagnostics = self._runtime_diagnostics(qt_context)
        logger.info(
            "OpenGL viewport initialized render_size=%s dpr=%.3f vendor=%s renderer=%s version=%s glsl=%s version_code=%s default_framebuffer=%s is_opengles=%s actual_format=%s requested_format=%s",
            self.current_render_size(),
            max(float(self.devicePixelRatioF()), 1.0),
            diagnostics["vendor"],
            diagnostics["renderer"],
            diagnostics["version"],
            diagnostics["glsl"],
            diagnostics["version_code"],
            diagnostics["default_framebuffer"],
            diagnostics["is_opengles"],
            diagnostics["actual_format"],
            diagnostics["requested_format"],
        )
        logger.debug(
            "initializeGL end current_context_present=%s moderngl_ctx_present=%s",
            QtGui.QOpenGLContext.currentContext() is not None,
            self._ctx is not None,
        )
        self.host_ready.emit()

    def paintGL(self) -> None:
        if self._ctx is None:
            return
        try:
            target = self._current_widget_target()
            self._prepare_render_target("paintGL", target=target)
            if self._scene is None:
                target.clear_color(*self._empty_clear_color())
                self._ctx.finish()
                self._paint_error_signature = None
                return
            logger.debug(
                "paintGL rendering scene_type=%s framebuffer=%s render_size=%s",
                type(self._scene).__name__,
                getattr(target.fbo, "glo", None),
                target.size,
            )
            self._scene.render_to(target)
            self._ctx.finish()
            self._paint_error_signature = None
        except Exception as exc:
            signature = (type(exc).__name__, str(exc))
            if signature != self._paint_error_signature:
                logger.exception(
                    "Viewport paint failed scene_type=%s render_size=%s",
                    type(self._scene).__name__ if self._scene is not None else None,
                    self.current_render_size(),
                )
                self.runtime_error.emit("Viewport rendering failed. See Log for details.")
                self._paint_error_signature = signature
            try:
                target = self._current_widget_target()
                self._prepare_render_target("paintGL-error", target=target)
                target.clear_color(*self._empty_clear_color())
                self._ctx.finish()
            except Exception:
                logger.debug("Viewport clear after paint failure also failed.", exc_info=True)

    def _event_point(self, event) -> tuple[float, float]:
        scale = max(self.height() / 2.0, 1.0)
        return event.position().x() / scale, event.position().y() / scale

    def mousePressEvent(self, event) -> None:
        self._mouse_pointer = self._event_point(event)
        self._drag_changed = False
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event) -> None:
        if self._session is None:
            return super().mouseMoveEvent(event)
        buttons = event.buttons()
        if not buttons:
            self._mouse_pointer = None
            self._drag_changed = False
            return super().mouseMoveEvent(event)

        point = self._event_point(event)
        if self._mouse_pointer is None:
            self._mouse_pointer = point
            return super().mouseMoveEvent(event)

        dx = point[0] - self._mouse_pointer[0]
        dy = point[1] - self._mouse_pointer[1]
        modifiers = event.modifiers()
        if buttons & QtCore.Qt.MouseButton.RightButton or (
            modifiers & QtCore.Qt.KeyboardModifier.ShiftModifier
        ):
            event_type = "pan"
        else:
            event_type = "rotate"
        changed = self._session.apply_camera_delta(
            event_type=event_type,
            dx=dx,
            dy=dy,
        )
        self._mouse_pointer = point
        if changed:
            self._drag_changed = True
            self.update()
            self.session_state_changed.emit(False)
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event) -> None:
        if self._session is not None and self._drag_changed:
            changed = self._session.commit_camera_interaction()
            self.update()
            self.session_state_changed.emit(bool(changed))
        self._mouse_pointer = None
        self._drag_changed = False
        super().mouseReleaseEvent(event)

    def wheelEvent(self, event) -> None:
        if self._session is None:
            return super().wheelEvent(event)
        delta = event.angleDelta()
        amount = delta.y() if abs(delta.y()) >= abs(delta.x()) else delta.x()
        if not amount:
            return super().wheelEvent(event)
        if event.modifiers() & QtCore.Qt.KeyboardModifier.ShiftModifier:
            changed = self._session.apply_camera_delta(
                event_type="roll",
                dx=float(amount) / 120.0,
            )
        else:
            changed = self._session.apply_camera_delta(
                event_type="zoom",
                wheel=float(amount) / 120.0,
            )
        if changed:
            self._session.commit_camera_interaction()
            self.update()
            self.session_state_changed.emit(True)
        super().wheelEvent(event)

    def apply_camera_button(self, action: str) -> bool:
        if self._session is None:
            return False
        changed = self._session.apply_camera_button(action)
        if changed:
            self.update()
            self.session_state_changed.emit(True)
        return changed


QtViewportWindow = QtViewport


__all__ = [
    "QtOpenGLLoader",
    "QtViewport",
    "QtViewportHost",
    "QtViewportWindow",
]



import moderngl as mgl
import numpy as np

from .render import *
from .buffer import Float3DBuffer, IntBuffer
from ..core import NumericProperty

######################################################################################################


class ChargeRender(StateRender):
    VERTEX = """
                #version 330

                const float M_PI = 3.1415926535897932384626433832795;

                uniform mat4 u_view;
                uniform mat4 u_projection;

                in vec3 in_center; // center of atom
                in vec3 in_spin; // spin direction

                out vec3 v_pos; // vertex position in world coordiantes
                out vec3 v_spin;

                mat3 rotmat(vec3 z) {
                    vec3 a = cross(z, vec3(1.0,0.0,0.0));
                    vec3 b = cross(z, vec3(0.0,1.0,0.0));
                    vec3 c = (length(a)>length(b)) ? a : b;
                    vec3 x = normalize(c);
                    vec3 y = cross(x, z);
                    return mat3(x, y, z);
                }

                void main() {
                    float len = length(in_spin);
                    vec3 spin = in_spin/len;
                    mat3 mesh = rotmat(spin);
                    vec3 v_pos = in_center;
                    gl_Position = u_projection * u_view * vec4(v_pos, 1.0);

                    v_spin = spin;
                } 
            """
    GEOMETRY = None
    FRAGMENT = """
                #version 330

                in vec4 v_color;

                out vec4 f_color;

                void main() {
                    f_color = v_color;
                }
            """

    def __init__(
        self, light=(1, 2, -3), coordinatesBuffer=None, maskBuffer=None, **kwargs
    ):
        super().__init__(**kwargs)

        self.u_view = self._prog["u_view"]
        self.u_projection = self._prog["u_projection"]
        self.u_camera = self._prog["u_camera"]
        self.u_light_direction = self._prog["u_light_direction"]
        self.u_use_mask = self._prog["u_use_mask"]
        self.u_min_size = self._prog["u_min_size"]
        self.u_max_size = self._prog["u_max_size"]
        self.u_power_size = self._prog["u_power_size"]
        self.u_min_threshold = self._prog["u_min_threshold"]
        self.u_size = self._prog["u_size"]
        self.u_clip_min = self._prog["u_clip_min"]
        self.u_clip_max = self._prog["u_clip_max"]
        self.u_clip_invert = self._prog["u_clip_invert"]

        self._light = np.array(light, dtype=np.float32)
        self._light /= np.linalg.norm(self._light)

        self._coordinatesBuffer = coordinatesBuffer
        self._coordinatesBuffer.upload(self._system.spin_positions())

        self._maskBuffer = maskBuffer

        self.mesh = ConicMesh(mglctx=self._ctx)

        self._vao = self._ctx.vertex_array(
            self._prog,
            [
                (self.mesh.pos.vbo, "3f", "in_pos"),
                (self.mesh.normal.vbo, "3f", "in_normal"),
                (self._state.vbo, "3f/i", "in_spin"),
                (self._coordinatesBuffer.vbo, "3f/i", "in_center"),
                (self._maskBuffer.vbo, "3f/i", "in_mask"),
            ],
            self.mesh.idx.ibo,
        )

        self.use_mask = False
        self.base_size = 1.0
        self.min_size = 0.0
        self.max_size = 1.5
        self.power_size = 1.0
        self.min_threshold = 0.5
        self.size = self._system.size + (1,) * (3 - self._system.dim)
        self.clip_min = (0, 0, 0)
        self.clip_max = tuple(self.size)
        self.clip_invert = False

    def render(
        self, scales=None, modelview=None, projection=None, camera_position=None
    ):
        self.u_view.value = tuple(modelview.T.flatten())
        self.u_projection.value = tuple(projection.T.flatten())
        self.u_camera.value = tuple(camera_position)
        self.u_light_direction.value = tuple(self._light)
        self.u_use_mask.value = self.use_mask
        self.u_min_size.value = self.min_size * self.base_size
        self.u_max_size.value = self.max_size * self.base_size
        self.u_power_size.value = self.power_size
        self.u_min_threshold.value = self.min_threshold
        self.u_size.value = tuple(self.size)
        self.u_clip_min.value = tuple(self.clip_min)
        self.u_clip_max.value = tuple(self.clip_max)
        self.u_clip_invert.value = self.clip_invert

        self._vao.render(
            mode=mgl.TRIANGLES,
            instances=self._system.number_of_spins,
            vertices=-1,
            first=0,
        )

    def list_properties(self):
        return [
            NumericProperty(
                self, "min_size", title="Min. size", min=0.0, max=2.0, count=30
            ),
            NumericProperty(
                self, "max_size", title="Max. size", min=0.0, max=2.0, count=30
            ),
            NumericProperty(
                self, "power_size", title="Steepness", min=0.0, max=2.0, count=30
            ),
            NumericProperty(
                self,
                "min_threshold",
                title="Min. threshold",
                min=0.0,
                max=1.0,
                count=50,
            ),
        ]



from .log import log, configLog
from .property import (
    Property,
    ListProperty,
    NumericProperty,
    BooleanProperty,
    PropertiesCollection,
    UpdateMask,
)
from .system import System
from .colors import vector_to_rgb
from .topology import hopf_charge

__all__ = [
    "log",
    "configLog",
    "Property",
    "ListProperty",
    "NumericProperty",
    "BooleanProperty",
    "PropertiesCollection",
    "UpdateMask",
    "System",
    "vector_to_rgb",
    "hopf_charge",
]



"""Qt bridge for viewer logging."""

from __future__ import annotations

import logging

from PySide6 import QtCore


class _QtLogRelay(QtCore.QObject):
    """QObject wrapper used to forward logging records across threads."""

    message_logged = QtCore.Signal(str, int)


class QtLogHandler(logging.Handler):
    """Logging handler that forwards formatted records into Qt widgets."""

    def __init__(self, *, level: int = logging.NOTSET):
        super().__init__(level=level)
        self.relay = _QtLogRelay()
        self.setFormatter(
            logging.Formatter(
                fmt="%(asctime)s %(levelname)s %(name)s: %(message)s",
                datefmt="%H:%M:%S",
            )
        )

    def emit(self, record: logging.LogRecord) -> None:
        try:
            message = self.format(record)
        except Exception:
            self.handleError(record)
            return
        self.relay.message_logged.emit(message, int(record.levelno))


__all__ = ["QtLogHandler"]



"""Main window for the `topovec view` desktop application."""

from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Any

import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets

from ...mgl import SCENES
from ...core import UpdateMask
from ..document import StateEntry, ViewerDocument, load_viewer_document
from ..residency import SceneResidencyManager
from ..session import ViewerSession
from .logging import QtLogHandler
from .theme import (
    THEME_MODE_DARK,
    THEME_MODE_LIGHT,
    THEME_MODE_SYSTEM,
    THEME_SETTING_KEY,
    apply_theme_mode,
    normalize_theme_mode,
)
from .viewport import QtViewport
from .widgets import ExportImageDialog, LogPanel, PropertyPanel, SnippetPanel

logger = logging.getLogger(__name__)


def _scene_supports_system(scene_cls: Any, system: Any) -> bool:
    supports = getattr(scene_cls, "supports_system", None)
    if not callable(supports):
        return True
    return bool(supports(system))


class TvViewWindow(QtWidgets.QMainWindow):
    """Single-window multi-document desktop viewer."""

    _WINDOW_STATE_VERSION = 4
    _RECENT_FILES_KEY = "recent_files"
    _MAX_RECENT_FILES = 10
    _ICON_THEME_FALLBACKS = ("Adwaita", "hicolor")
    _DEFAULT_THEME_MODE = THEME_MODE_SYSTEM

    def __init__(
        self,
        parent: QtWidgets.QWidget | None = None,
        *,
        max_live_scenes: int = 2,
    ):
        super().__init__(parent)
        logger.debug("Initializing TvViewWindow")
        self.setWindowTitle("topovec view")
        self.resize(1400, 900)

        self._settings = QtCore.QSettings("topovec", "topovec.view")
        self._documents: list[ViewerDocument] = []
        self._current_document: ViewerDocument | None = None
        self._current_entry: StateEntry | None = None
        self._pending_entry: tuple[ViewerDocument, StateEntry] | None = None
        self._session: ViewerSession | None = None
        self._scene_combo_updating = False
        self._preferred_scene_name = "Layer"
        self._status_override: str | None = None
        self._log_handler: QtLogHandler | None = None
        self._max_live_scenes = int(max_live_scenes)
        self._theme_action_group: QtGui.QActionGroup | None = None
        self._theme_actions: dict[str, QtGui.QAction] = {}
        self._theme_mode = self._load_theme_mode()
        self._theme_spec = None
        self._dock_area_actions: dict[QtCore.Qt.DockWidgetArea, QtGui.QAction] = {}
        self._dock_area_recomputing = False

        logger.debug("Creating central QtViewport")
        self.viewport = QtViewport(self)
        logger.debug("Applying QtViewport surface format")
        self.viewport.setFormat(QtViewport.create_surface_format())
        self.viewport.host_ready.connect(self._on_viewport_ready)
        self.viewport.session_state_changed.connect(self._on_session_state_changed)
        self.viewport.runtime_error.connect(self._on_viewport_runtime_error)
        self.setCentralWidget(self.viewport)
        self._scene_residency = SceneResidencyManager(
            self.viewport.host,
            scene_registry_loader=lambda: dict(SCENES),
            session_factory=ViewerSession,
            max_live_scenes=self._max_live_scenes,
        )

        logger.debug("Building topovec view docks")
        self._build_docks()
        self._apply_theme(self._theme_mode, persist=False)
        logger.debug("Installing topovec view GUI log handler")
        self._install_log_handler()
        logger.debug("Building topovec view actions and toolbar")
        self._build_actions()
        self._build_toolbar()
        logger.debug("Restoring topovec view window settings")
        restored_layout = self._restore_settings()
        if not restored_layout:
            self._apply_default_layout()
        self._sync_dock_area_actions()
        self._refresh_ui()
        style_hints = QtWidgets.QApplication.instance().styleHints()
        style_hints.colorSchemeChanged.connect(self._on_system_color_scheme_changed)
        logger.debug("TvViewWindow initialization complete")

    @property
    def session(self) -> ViewerSession | None:
        return self._session

    def _build_docks(self) -> None:
        self.sources_tree = QtWidgets.QTreeWidget()
        self.sources_tree.setHeaderLabels(["Source"])
        self.sources_tree.itemSelectionChanged.connect(self._on_tree_selection_changed)

        self.sources_dock = QtWidgets.QDockWidget("Data Sources", self)
        self.sources_dock.setObjectName("topovec.view.sources")
        self.sources_dock.setWidget(self.sources_tree)
        self.addDockWidget(QtCore.Qt.DockWidgetArea.LeftDockWidgetArea, self.sources_dock)

        self.properties_panel = PropertyPanel(self)
        self.properties_panel.property_edited.connect(self._on_property_edited)
        self.properties_dock = QtWidgets.QDockWidget("Properties", self)
        self.properties_dock.setObjectName("topovec.view.properties")
        self.properties_dock.setWidget(self.properties_panel)
        self.addDockWidget(
            QtCore.Qt.DockWidgetArea.RightDockWidgetArea,
            self.properties_dock,
        )

        self.snippet_panel = SnippetPanel(self)
        self.snippet_dock = QtWidgets.QDockWidget("Python Snippet", self)
        self.snippet_dock.setObjectName("topovec.view.snippet")
        self.snippet_dock.setWidget(self.snippet_panel)
        self.snippet_dock.visibilityChanged.connect(self._on_snippet_visibility_changed)
        self.addDockWidget(
            QtCore.Qt.DockWidgetArea.BottomDockWidgetArea,
            self.snippet_dock,
        )

        self.log_panel = LogPanel(self)
        self.log_panel.record_appended.connect(self._on_log_record_appended)
        self.log_dock = QtWidgets.QDockWidget("Log", self)
        self.log_dock.setObjectName("topovec.view.log")
        self.log_dock.setWidget(self.log_panel)
        self.addDockWidget(
            QtCore.Qt.DockWidgetArea.BottomDockWidgetArea,
            self.log_dock,
        )
        self._arrange_bottom_output_docks()
        self._ensure_dock_area_signal_connections()

    def _arrange_bottom_output_docks(self) -> None:
        for dock in (self.snippet_dock, self.log_dock):
            dock.setFloating(False)
            self.addDockWidget(QtCore.Qt.DockWidgetArea.BottomDockWidgetArea, dock)
        self.tabifyDockWidget(self.snippet_dock, self.log_dock)
        self.log_dock.raise_()

    def _install_log_handler(self) -> None:
        if self._log_handler is not None:
            return
        self._log_handler = QtLogHandler()
        self._log_handler.relay.message_logged.connect(
            self.log_panel.append_message,
            QtCore.Qt.ConnectionType.QueuedConnection,
        )
        logging.getLogger("topovec").addHandler(self._log_handler)

    def _uninstall_log_handler(self) -> None:
        if self._log_handler is None:
            return
        logging.getLogger("topovec").removeHandler(self._log_handler)
        self._log_handler.close()
        self._log_handler = None

    @classmethod
    def _candidate_icon_search_paths(cls) -> list[str]:
        candidates = list(QtGui.QIcon.themeSearchPaths())
        xdg_data_dirs = os.environ.get("XDG_DATA_DIRS", "/usr/local/share:/usr/share")
        for base in xdg_data_dirs.split(":"):
            if not base:
                continue
            candidates.append(os.fspath(Path(base) / "icons"))
        candidates.append("/usr/share/icons")
        candidates.append("/usr/local/share/icons")
        candidates.append(os.fspath(Path.home() / ".icons"))

        paths = []
        seen = set()
        for path in candidates:
            normalized = os.fspath(Path(path).expanduser())
            if normalized in seen or not Path(normalized).is_dir():
                continue
            seen.add(normalized)
            paths.append(normalized)
        return paths

    @classmethod
    def _available_icon_theme(cls, search_paths: list[str]) -> str | None:
        for theme_name in cls._ICON_THEME_FALLBACKS:
            for root in search_paths:
                if (Path(root) / theme_name / "index.theme").is_file():
                    return theme_name
        return None

    @classmethod
    def _ensure_theme_icon_support(cls) -> None:
        search_paths = cls._candidate_icon_search_paths()
        if search_paths:
            current_paths = list(QtGui.QIcon.themeSearchPaths())
            if current_paths != search_paths:
                QtGui.QIcon.setThemeSearchPaths(search_paths)
        if QtGui.QIcon.themeName() or QtGui.QIcon.fallbackThemeName():
            return
        fallback_theme = cls._available_icon_theme(search_paths)
        if fallback_theme:
            QtGui.QIcon.setFallbackThemeName(fallback_theme)

    def _icon_from_theme(
        self,
        names: tuple[str, ...],
        fallback: QtWidgets.QStyle.StandardPixmap,
    ) -> QtGui.QIcon:
        style = self.style()
        for name in names:
            icon = QtGui.QIcon.fromTheme(name)
            if not icon.isNull():
                return icon
        for fallback_pixmap in (
            fallback,
            QtWidgets.QStyle.StandardPixmap.SP_FileIcon,
            QtWidgets.QStyle.StandardPixmap.SP_DialogOpenButton,
        ):
            icon = style.standardIcon(fallback_pixmap)
            if not icon.isNull():
                return icon
        return QtGui.QIcon()

    def _configure_action_icons(self) -> None:
        self._ensure_theme_icon_support()
        self.open_action.setIcon(
            self._icon_from_theme(
                ("document-open", "folder-open"),
                QtWidgets.QStyle.StandardPixmap.SP_DialogOpenButton,
            )
        )
        self.save_image_action.setIcon(
            self._icon_from_theme(
                ("document-save-as", "document-save"),
                QtWidgets.QStyle.StandardPixmap.SP_DialogSaveButton,
            )
        )
        self.reset_camera_action.setIcon(
            self._icon_from_theme(
                ("go-home", "view-refresh"),
                QtWidgets.QStyle.StandardPixmap.SP_RestoreDefaultsButton,
            )
        )
        self.snap_camera_action.setIcon(
            self._icon_from_theme(
                (
                    "orientation-landscape-symbolic",
                    "camera-switch-symbolic",
                    "object-select-symbolic",
                ),
                QtWidgets.QStyle.StandardPixmap.SP_ArrowUp,
            )
        )
        self.toggle_projection_action.setIcon(
            self._icon_from_theme(
                ("view-grid-symbolic", "view-app-grid-symbolic", "view-grid"),
                QtWidgets.QStyle.StandardPixmap.SP_FileDialogDetailedView,
            )
        )

    def _build_actions(self) -> None:
        self.open_action = QtGui.QAction("Open...", self)
        self.open_action.setShortcut(QtGui.QKeySequence.StandardKey.Open)
        self.open_action.triggered.connect(self.open_files)

        self.open_recent_menu = QtWidgets.QMenu("Open Recent", self)
        self.open_recent_menu.aboutToShow.connect(self._refresh_recent_files_menu)
        self.clear_recent_files_action = QtGui.QAction("Clear Recent Files", self)
        self.clear_recent_files_action.triggered.connect(self.clear_recent_files)

        self.save_image_action = QtGui.QAction("Save Image...", self)
        self.save_image_action.setShortcut(QtGui.QKeySequence.StandardKey.SaveAs)
        self.save_image_action.triggered.connect(self.save_image)

        self.close_source_action = QtGui.QAction("Close Selected Source", self)
        self.close_source_action.triggered.connect(self.close_selected_source)

        self.exit_action = QtGui.QAction("Exit", self)
        self.exit_action.setShortcut(QtGui.QKeySequence.StandardKey.Quit)
        self.exit_action.triggered.connect(self.close)

        self.reset_camera_action = QtGui.QAction("Reset Camera", self)
        self.reset_camera_action.triggered.connect(
            lambda: self.viewport.apply_camera_button("reset")
        )

        self.snap_camera_action = QtGui.QAction("Snap Camera", self)
        self.snap_camera_action.triggered.connect(
            lambda: self.viewport.apply_camera_button("snap")
        )

        self.toggle_projection_action = QtGui.QAction("Toggle Projection", self)
        self.toggle_projection_action.triggered.connect(
            lambda: self.viewport.apply_camera_button("toggle_perspective")
        )

        self.reset_layout_action = QtGui.QAction("Reset Layout", self)
        self.reset_layout_action.triggered.connect(self.reset_layout)
        self._configure_action_icons()
        self.left_panel_action = QtGui.QAction("Left Panel", self)
        self.left_panel_action.setCheckable(True)
        self.left_panel_action.triggered.connect(
            lambda checked=False: self._set_area_visible(
                QtCore.Qt.DockWidgetArea.LeftDockWidgetArea,
                bool(checked),
            )
        )
        self.bottom_panel_action = QtGui.QAction("Bottom Panel", self)
        self.bottom_panel_action.setCheckable(True)
        self.bottom_panel_action.triggered.connect(
            lambda checked=False: self._set_area_visible(
                QtCore.Qt.DockWidgetArea.BottomDockWidgetArea,
                bool(checked),
            )
        )
        self.right_panel_action = QtGui.QAction("Right Panel", self)
        self.right_panel_action.setCheckable(True)
        self.right_panel_action.triggered.connect(
            lambda checked=False: self._set_area_visible(
                QtCore.Qt.DockWidgetArea.RightDockWidgetArea,
                bool(checked),
            )
        )
        self._dock_area_actions = {
            QtCore.Qt.DockWidgetArea.LeftDockWidgetArea: self.left_panel_action,
            QtCore.Qt.DockWidgetArea.BottomDockWidgetArea: self.bottom_panel_action,
            QtCore.Qt.DockWidgetArea.RightDockWidgetArea: self.right_panel_action,
        }
        self.left_panel_action.setIcon(
            self._icon_from_theme(
                ("sidebar-show-symbolic",),
                QtWidgets.QStyle.StandardPixmap.SP_ArrowLeft,
            )
        )
        self.bottom_panel_action.setIcon(
            self._icon_from_theme(
                ("pan-down-symbolic", "go-down-symbolic"),
                QtWidgets.QStyle.StandardPixmap.SP_ArrowDown,
            )
        )
        self.right_panel_action.setIcon(
            self._icon_from_theme(
                ("sidebar-show-right-symbolic",),
                QtWidgets.QStyle.StandardPixmap.SP_ArrowRight,
            )
        )

        file_menu = self.menuBar().addMenu("&File")
        file_menu.addAction(self.open_action)
        file_menu.addMenu(self.open_recent_menu)
        file_menu.addAction(self.save_image_action)
        file_menu.addAction(self.close_source_action)
        file_menu.addSeparator()
        file_menu.addAction(self.exit_action)

        view_menu = self.menuBar().addMenu("&View")
        view_menu.addAction(self.reset_camera_action)
        view_menu.addAction(self.snap_camera_action)
        view_menu.addAction(self.toggle_projection_action)
        view_menu.addSeparator()
        view_menu.addAction(self.left_panel_action)
        view_menu.addAction(self.bottom_panel_action)
        view_menu.addAction(self.right_panel_action)
        view_menu.addSeparator()
        self._theme_action_group = QtGui.QActionGroup(self)
        self._theme_action_group.setExclusive(True)
        theme_menu = view_menu.addMenu("Theme")
        self._theme_actions = {}
        for mode, label in (
            (THEME_MODE_SYSTEM, "System"),
            (THEME_MODE_DARK, "Dark"),
            (THEME_MODE_LIGHT, "Light"),
        ):
            action = QtGui.QAction(label, self)
            action.setCheckable(True)
            action.triggered.connect(
                lambda checked=False, selected_mode=mode: self._on_theme_action_triggered(
                    selected_mode, checked
                )
            )
            self._theme_action_group.addAction(action)
            theme_menu.addAction(action)
            self._theme_actions[mode] = action
        view_menu.addAction(self.reset_layout_action)
        view_menu.addSeparator()
        view_menu.addAction(self.sources_dock.toggleViewAction())
        view_menu.addAction(self.properties_dock.toggleViewAction())
        view_menu.addAction(self.snippet_dock.toggleViewAction())
        view_menu.addAction(self.log_dock.toggleViewAction())
        self._sync_theme_actions()
        self._refresh_recent_files_menu()

    def _build_toolbar(self) -> None:
        toolbar = self.addToolBar("Main")
        toolbar.setObjectName("topovec.view.toolbar")
        toolbar.setToolButtonStyle(QtCore.Qt.ToolButtonStyle.ToolButtonIconOnly)
        toolbar.setIconSize(QtCore.QSize(18, 18))
        toolbar.addAction(self.open_action)
        toolbar.addAction(self.save_image_action)
        toolbar.addSeparator()
        toolbar.addAction(self.left_panel_action)
        toolbar.addAction(self.bottom_panel_action)
        toolbar.addAction(self.right_panel_action)
        toolbar.addSeparator()
        toolbar.addAction(self.reset_camera_action)
        toolbar.addAction(self.snap_camera_action)
        toolbar.addAction(self.toggle_projection_action)
        spacer = QtWidgets.QWidget()
        spacer.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Preferred,
        )
        toolbar.addWidget(spacer)
        toolbar.addWidget(QtWidgets.QLabel("Scene"))
        self.scene_combo = QtWidgets.QComboBox()
        self.scene_combo.currentTextChanged.connect(self._on_scene_changed)
        toolbar.addWidget(self.scene_combo)
        self._sync_dock_area_actions()

    def _restore_settings(self) -> bool:
        geometry = self._settings.value("geometry")
        if geometry is not None:
            self.restoreGeometry(geometry)
        state = self._settings.value("window_state")
        state_version = int(self._settings.value("window_state_version", 0) or 0)
        if state is not None and state_version == self._WINDOW_STATE_VERSION:
            self.restoreState(state)
            return True
        if state is not None:
            logger.info(
                "Ignoring incompatible topovec view window_state version stored=%s expected=%s",
                state_version,
                self._WINDOW_STATE_VERSION,
            )
        return False

    def _store_settings(self) -> None:
        self._settings.setValue("geometry", self.saveGeometry())
        self._settings.setValue("window_state", self.saveState())
        self._settings.setValue("window_state_version", self._WINDOW_STATE_VERSION)

    def _load_theme_mode(self) -> str:
        return normalize_theme_mode(
            self._settings.value(THEME_SETTING_KEY, self._DEFAULT_THEME_MODE)
        )

    def _sync_theme_actions(self) -> None:
        for mode, action in self._theme_actions.items():
            action.setChecked(mode == self._theme_mode)

    def _apply_theme(self, mode: str, *, persist: bool) -> None:
        app = QtWidgets.QApplication.instance()
        self._theme_mode = normalize_theme_mode(mode)
        self._theme_spec = apply_theme_mode(app, self._theme_mode)
        self.properties_panel.apply_theme(self._theme_spec)
        self.viewport.apply_theme(self._theme_spec)
        self._sync_theme_actions()
        if persist:
            self._settings.setValue(THEME_SETTING_KEY, self._theme_mode)

    def _dock_widgets(self) -> list[QtWidgets.QDockWidget]:
        return list(
            self.findChildren(
                QtWidgets.QDockWidget,
                options=QtCore.Qt.FindChildOption.FindDirectChildrenOnly,
            )
        )

    def _connect_dock_area_signals(self, dock: QtWidgets.QDockWidget) -> None:
        if bool(dock.property("_topovecViewDockAreaSignalsConnected")):
            return
        dock.visibilityChanged.connect(self._on_dock_visibility_changed)
        dock.topLevelChanged.connect(self._on_dock_top_level_changed)
        dock.dockLocationChanged.connect(self._on_dock_location_changed)
        dock.setProperty("_topovecViewDockAreaSignalsConnected", True)

    def _ensure_dock_area_signal_connections(self) -> None:
        for dock in self._dock_widgets():
            self._connect_dock_area_signals(dock)

    def _docks_in_area(self, area: QtCore.Qt.DockWidgetArea) -> list[QtWidgets.QDockWidget]:
        self._ensure_dock_area_signal_connections()
        docks = []
        for dock in self._dock_widgets():
            if dock.isFloating():
                continue
            if self.dockWidgetArea(dock) == area:
                docks.append(dock)
        return docks

    def _set_area_visible(
        self,
        area: QtCore.Qt.DockWidgetArea,
        visible: bool,
    ) -> None:
        docks = self._docks_in_area(area)
        for dock in docks:
            dock.setVisible(bool(visible))
        self._sync_dock_area_actions()

    def _sync_dock_area_actions(self) -> None:
        if self._dock_area_recomputing:
            return
        self._dock_area_recomputing = True
        try:
            for area, action in self._dock_area_actions.items():
                visible = any(not dock.isHidden() for dock in self._docks_in_area(area))
                blocker = QtCore.QSignalBlocker(action)
                action.setChecked(bool(visible))
                del blocker
        finally:
            self._dock_area_recomputing = False

    def _on_dock_visibility_changed(self, _visible: bool) -> None:
        self._sync_dock_area_actions()

    def _on_dock_top_level_changed(self, _floating: bool) -> None:
        self._sync_dock_area_actions()

    def _on_dock_location_changed(self, _area: QtCore.Qt.DockWidgetArea) -> None:
        self._sync_dock_area_actions()

    def childEvent(self, event: QtCore.QChildEvent) -> None:
        super().childEvent(event)
        if event.added():
            child = event.child()
            if isinstance(child, QtWidgets.QDockWidget):
                self._connect_dock_area_signals(child)
                self._sync_dock_area_actions()

    def _on_theme_action_triggered(self, mode: str, checked: bool) -> None:
        if not checked:
            return
        self._apply_theme(mode, persist=True)

    def _on_system_color_scheme_changed(self, _scheme) -> None:
        if self._theme_mode != THEME_MODE_SYSTEM:
            return
        self._apply_theme(self._theme_mode, persist=False)

    def _recent_file_paths(self) -> list[Path]:
        raw = self._settings.value(self._RECENT_FILES_KEY, [])
        if raw is None:
            return []
        if isinstance(raw, str):
            values = [raw]
        elif isinstance(raw, (list, tuple)):
            values = [str(value) for value in raw]
        else:
            values = [str(raw)]

        paths = []
        seen = set()
        for value in values:
            try:
                path = Path(value).expanduser().resolve()
            except Exception:
                logger.debug("Ignoring unreadable recent file entry value=%r", value)
                continue
            key = os.fspath(path)
            if key in seen:
                continue
            seen.add(key)
            paths.append(path)
        return paths

    def _store_recent_file_paths(self, paths: list[Path]) -> None:
        self._settings.setValue(
            self._RECENT_FILES_KEY,
            [os.fspath(path) for path in paths[: self._MAX_RECENT_FILES]],
        )

    def _prune_recent_file_paths(self) -> list[Path]:
        paths = self._recent_file_paths()
        existing = [path for path in paths if path.exists()]
        if existing != paths:
            logger.info(
                "Pruned stale recent files removed=%s kept=%s",
                len(paths) - len(existing),
                len(existing),
            )
            self._store_recent_file_paths(existing)
        return existing

    def _remember_recent_paths(self, paths: list[Path]) -> None:
        remembered = self._recent_file_paths()
        merged = []
        seen = set()
        for path in [*paths, *remembered]:
            key = os.fspath(path)
            if key in seen:
                continue
            seen.add(key)
            merged.append(path)
            if len(merged) >= self._MAX_RECENT_FILES:
                break
        self._store_recent_file_paths(merged)
        self._refresh_recent_files_menu()

    def clear_recent_files(self) -> None:
        logger.info("Clearing remembered recent files")
        self._settings.remove(self._RECENT_FILES_KEY)
        self._refresh_recent_files_menu()

    def _recent_file_labels(self, paths: list[Path]) -> dict[Path, str]:
        counts = {}
        for path in paths:
            counts[path.name] = counts.get(path.name, 0) + 1
        labels = {}
        for path in paths:
            if counts[path.name] == 1:
                labels[path] = path.name
            else:
                labels[path] = f"{path.name} — {path.parent}"
        return labels

    def _refresh_recent_files_menu(self) -> None:
        paths = self._prune_recent_file_paths()
        self.open_recent_menu.clear()
        labels = self._recent_file_labels(paths)
        for path in paths:
            action = self.open_recent_menu.addAction(labels[path])
            action.setToolTip(os.fspath(path))
            action.setStatusTip(os.fspath(path))
            action.triggered.connect(
                lambda _checked=False, path_str=os.fspath(path): self.open_recent_file(path_str)
            )
        if paths:
            self.open_recent_menu.addSeparator()
        self.clear_recent_files_action.setEnabled(bool(paths))
        self.open_recent_menu.addAction(self.clear_recent_files_action)
        self.open_recent_menu.menuAction().setEnabled(bool(paths))

    def _apply_default_dock_sizes(self) -> None:
        self.resizeDocks(
            [self.sources_dock, self.properties_dock],
            [240, 320],
            QtCore.Qt.Orientation.Horizontal,
        )

    def _apply_default_layout(self) -> None:
        for dock, area in (
            (self.sources_dock, QtCore.Qt.DockWidgetArea.LeftDockWidgetArea),
            (self.properties_dock, QtCore.Qt.DockWidgetArea.RightDockWidgetArea),
            (self.snippet_dock, QtCore.Qt.DockWidgetArea.BottomDockWidgetArea),
            (self.log_dock, QtCore.Qt.DockWidgetArea.BottomDockWidgetArea),
        ):
            dock.setFloating(False)
            self.addDockWidget(area, dock)
            dock.show()
        self._arrange_bottom_output_docks()
        self.snippet_dock.hide()
        self.log_dock.hide()
        self._apply_default_dock_sizes()
        self._sync_dock_area_actions()

    def reset_layout(self) -> None:
        logger.info("Resetting topovec view dock layout to defaults")
        self._settings.remove("window_state")
        self._settings.remove("window_state_version")
        self._apply_default_layout()

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self._store_settings()
        try:
            self.viewport.makeCurrent()
            self._scene_residency.release()
        except Exception:
            logger.exception("Failed to release residency-managed scenes during shutdown")
        finally:
            try:
                self.viewport.doneCurrent()
            except Exception:
                pass
        self.viewport.shutdown()
        self._uninstall_log_handler()
        super().closeEvent(event)

    def _focus_open_document(self, document: ViewerDocument) -> bool:
        for index in range(self.sources_tree.topLevelItemCount()):
            item = self.sources_tree.topLevelItem(index)
            payload = item.data(0, QtCore.Qt.ItemDataRole.UserRole)
            if not payload or payload[0] != "document" or payload[1] is not document:
                continue
            if item.childCount() > 0:
                self.sources_tree.setCurrentItem(item.child(0))
            else:
                self.sources_tree.setCurrentItem(item)
            return True
        return False

    def _open_document_by_path(self, path: Path) -> ViewerDocument | None:
        for document in self._documents:
            if document.path == path:
                return document
        return None

    def load_paths(self, paths: list[str] | tuple[str, ...]) -> list[ViewerDocument]:
        logger.info("Loading viewer documents count=%s", len(paths))
        first_state_item = None
        loaded_documents = []
        for raw_path in paths:
            try:
                document = load_viewer_document(raw_path)
            except Exception as exc:
                logger.exception("Failed to open viewer document path=%s", raw_path)
                self._set_status_override(
                    f"Failed to open {Path(raw_path).name}. See Log for details."
                )
                QtWidgets.QMessageBox.warning(
                    self,
                    "Open Failed",
                    f"Failed to open {raw_path!r}.\n\n{exc}",
                )
                continue
            logger.debug(
                "Loaded document path=%s format=%s entries=%s system_size=%s",
                document.path,
                document.format_name,
                len(document.entries),
                tuple(int(v) for v in getattr(document.system, "size", ())),
            )
            self._documents.append(document)
            loaded_documents.append(document)
            doc_item = QtWidgets.QTreeWidgetItem(
                [f"{document.label} [{document.format_name}]"]
            )
            doc_item.setData(0, QtCore.Qt.ItemDataRole.UserRole, ("document", document))
            self.sources_tree.addTopLevelItem(doc_item)
            for entry in document.entries:
                child = QtWidgets.QTreeWidgetItem([entry.label])
                child.setData(
                    0,
                    QtCore.Qt.ItemDataRole.UserRole,
                    ("state", document, entry),
                )
                doc_item.addChild(child)
                if first_state_item is None:
                    first_state_item = child
            doc_item.setExpanded(True)
        if loaded_documents:
            self._remember_recent_paths([document.path for document in loaded_documents])
        if first_state_item is not None and self.sources_tree.currentItem() is None:
            self.sources_tree.setCurrentItem(first_state_item)
        self._refresh_ui()
        return loaded_documents

    def open_files(self) -> None:
        filenames, _filter = QtWidgets.QFileDialog.getOpenFileNames(
            self,
            "Open Sources",
            "",
            "Viewer files (*.npz *.sad);;NPZ files (*.npz);;SAD files (*.sad);;All files (*)",
        )
        if filenames:
            self.load_paths(list(filenames))

    def open_recent_file(self, raw_path: str) -> None:
        path = Path(raw_path).expanduser().resolve()
        if not path.exists():
            logger.info("Recent file no longer exists path=%s", path)
            self._prune_recent_file_paths()
            self._set_status_override(f"Recent file is no longer available: {path.name}")
            return
        existing = self._open_document_by_path(path)
        if existing is not None:
            logger.info("Focusing already opened recent file path=%s", path)
            self._remember_recent_paths([path])
            self._focus_open_document(existing)
            self._status_override = None
            self._refresh_status_and_actions()
            return
        loaded_documents = self.load_paths([os.fspath(path)])
        if loaded_documents:
            self._focus_open_document(loaded_documents[0])

    def close_selected_source(self) -> None:
        item = self.sources_tree.currentItem()
        if item is None:
            return
        payload = item.data(0, QtCore.Qt.ItemDataRole.UserRole)
        if not payload:
            return
        if payload[0] == "state":
            document = payload[1]
            item = item.parent() or item
        else:
            document = payload[1]
        if document in self._documents:
            self._documents.remove(document)
        index = self.sources_tree.indexOfTopLevelItem(item)
        if index >= 0:
            self.sources_tree.takeTopLevelItem(index)
        if self._current_document is document:
            self._clear_session()
            if self.sources_tree.topLevelItemCount() > 0:
                first = self.sources_tree.topLevelItem(0)
                if first.childCount() > 0:
                    self.sources_tree.setCurrentItem(first.child(0))
        self._refresh_ui()

    def _clear_session(self) -> None:
        self._scene_residency.deactivate()
        self._session = None
        self._current_document = None
        self._current_entry = None
        self._status_override = None
        self.viewport.clear_session()

    def _request_viewport_redraw(self, reason: str) -> None:
        logger.debug("Requesting viewport redraw reason=%s", reason)
        self.viewport.update()

    def _on_viewport_ready(self) -> None:
        logger.debug(
            "Viewport ready pending_entry=%s",
            self._pending_entry is not None,
        )
        if self._pending_entry is not None:
            document, entry = self._pending_entry
            self._pending_entry = None
            self._activate_entry(document, entry)

    def _on_tree_selection_changed(self) -> None:
        item = self.sources_tree.currentItem()
        if item is None:
            return
        payload = item.data(0, QtCore.Qt.ItemDataRole.UserRole)
        if not payload or payload[0] != "state":
            return
        _kind, document, entry = payload
        logger.info(
            "Selected state document=%s entry=%s entry_index=%s",
            document.path,
            entry.label,
            entry.index,
        )
        self._activate_entry(document, entry)

    def _preferred_scene_for_system(self, system) -> str:
        preferred_names = [self._preferred_scene_name, "Layer", "Vectors", "Ball"]
        seen = set()
        for name in preferred_names:
            if name in seen or name not in SCENES:
                continue
            seen.add(name)
            if _scene_supports_system(SCENES[name], system):
                return name
        for name, scene_cls in SCENES.items():
            if _scene_supports_system(scene_cls, system):
                return name
        raise RuntimeError("No built-in scenes are compatible with this system.")

    def _activate_entry(self, document: ViewerDocument, entry: StateEntry) -> None:
        if not self.viewport.is_ready:
            logger.debug(
                "Viewport is not ready yet; deferring activation document=%s entry=%s",
                document.path,
                entry.label,
            )
            self._pending_entry = (document, entry)
            return
        logger.debug(
            "Activating entry begin document=%s format=%s entry=%s entry_index=%s system_size=%s viewport_ready=%s has_previous_session=%s",
            document.path,
            document.format_name,
            entry.label,
            entry.index,
            tuple(int(v) for v in getattr(document.system, "size", ())),
            self.viewport.is_ready,
            self._session is not None,
        )
        previous_session = self._session
        previous_scene = self.viewport.host.scene
        previous_document = self._current_document
        previous_entry = self._current_entry
        logger.debug(
            "Activation residency start max_live_scenes=%s previous_scene=%s target_system_type=%s",
            self._max_live_scenes,
            None if previous_session is None else previous_session.selected_scene,
            None if document.system is None else type(document.system).__name__,
        )
        explicit_redraw_requests = 0
        try:
            state = entry.load_state()
            logger.debug(
                "Loaded state for activation document=%s entry=%s shape=%s dtype=%s",
                document.path,
                entry.label,
                tuple(int(v) for v in getattr(state, "shape", ())),
                getattr(state, "dtype", None),
            )
            preferred_scene = self._preferred_scene_for_system(document.system)
            session = self._scene_residency.activate(
                system=document.system,
                state=state,
                preferred_scene_name=preferred_scene,
            )
            scene_name = session.selected_scene or preferred_scene
            logger.debug(
                "Activated residency-managed session mode=%s scene=%s document=%s entry=%s",
                self._scene_residency.last_activation_mode,
                scene_name,
                document.path,
                entry.label,
            )
        except Exception:
            logger.exception(
                "Failed to activate entry document=%s entry=%s entry_index=%s mode=%s",
                document.path,
                entry.label,
                entry.index,
                self._scene_residency.last_activation_mode,
            )
            self._restore_previous_session(
                previous_session,
                previous_scene,
                previous_document,
                previous_entry,
            )
            self._set_status_override(
                f"Failed to activate {document.label} :: {entry.label}. See Log for details."
            )
            self._refresh_ui(full=self._session is not None)
            return

        self._session = session
        self.viewport.set_session(self._session)
        self._current_document = document
        self._current_entry = entry
        self._preferred_scene_name = self._session.selected_scene or scene_name
        self._status_override = None
        logger.info(
            "Activated entry document=%s entry=%s scene=%s",
            document.path,
            entry.label,
            self._preferred_scene_name,
        )
        self._refresh_ui(full=True)
        explicit_redraw_requests += 1
        self._request_viewport_redraw("activate_entry")
        logger.debug(
            "Activation explicit viewport redraw_requests=%s document=%s entry=%s",
            explicit_redraw_requests,
            document.path,
            entry.label,
        )

    def _restore_previous_session(
        self,
        session: ViewerSession | None,
        scene: object,
        document: ViewerDocument | None,
        entry: StateEntry | None,
    ) -> None:
        if session is None or scene is None:
            self._clear_session()
            return
        try:
            self.viewport.host.attach_scene(scene)
            self._session = session
            self.viewport.set_session(session)
            self._current_document = document
            self._current_entry = entry
            self._request_viewport_redraw("restore_previous_session")
        except Exception:
            logger.exception("Failed to restore the previous viewer session after an activation error")
            self._clear_session()

    def _refresh_scene_combo(self) -> None:
        self._scene_combo_updating = True
        try:
            self.scene_combo.clear()
            if self._session is None:
                self.scene_combo.setEnabled(False)
                return
            self.scene_combo.addItems(self._session.available_scenes)
            current = self._session.selected_scene
            if current is not None:
                index = self.scene_combo.findText(current)
                if index >= 0:
                    self.scene_combo.setCurrentIndex(index)
            self.scene_combo.setEnabled(bool(self._session.available_scenes))
        finally:
            self._scene_combo_updating = False

    def _status_text(self) -> str:
        if self._status_override:
            return self._status_override
        if self._session is None:
            return "No state selected."
        projection = self._session.camera_snapshot()["projection_label"]
        scene = self._session.selected_scene or "Unknown scene"
        source = (
            f"{self._current_document.label} :: {self._current_entry.label}"
            if self._current_document is not None and self._current_entry is not None
            else "Unknown source"
        )
        return f"{source} :: {scene} :: {projection}"

    def _refresh_status_and_actions(self) -> None:
        self.statusBar().showMessage(self._status_text())
        has_session = self._session is not None
        self.save_image_action.setEnabled(has_session)
        self.close_source_action.setEnabled(self.sources_tree.currentItem() is not None)
        self.reset_camera_action.setEnabled(has_session)
        self.snap_camera_action.setEnabled(has_session)
        self.toggle_projection_action.setEnabled(has_session)

    def _refresh_summary(self) -> None:
        if self._session is None:
            self.snippet_panel.set_text("")
            return
        if self.snippet_dock.isVisible():
            self.snippet_panel.set_text(self._session.render_snippet())

    def _refresh_properties_panel(self, *, rebuild: bool) -> None:
        if self._session is None:
            self.properties_panel.set_snapshot(None, rebuild=True)
            return
        self.properties_panel.set_snapshot(
            self._session.properties_snapshot(),
            rebuild=bool(rebuild),
        )

    def _refresh_ui(self, *, full: bool = False) -> None:
        self._refresh_scene_combo()
        if full and self._session is not None:
            self._refresh_properties_panel(rebuild=True)
            self._refresh_summary()
        elif self._session is None:
            self._refresh_properties_panel(rebuild=True)
            self._refresh_summary()
        self._refresh_status_and_actions()

    def _on_session_state_changed(self, refresh_summary: bool) -> None:
        self._refresh_status_and_actions()
        if refresh_summary:
            self._refresh_summary()

    def _set_status_override(self, message: str | None) -> None:
        self._status_override = None if not message else str(message)
        self.statusBar().showMessage(self._status_text())

    def _on_viewport_runtime_error(self, message: str) -> None:
        self._set_status_override(message)

    def _on_snippet_visibility_changed(self, visible: bool) -> None:
        if visible and self._session is not None:
            self.snippet_panel.set_text(self._session.render_snippet())

    def _on_scene_changed(self, scene_name: str) -> None:
        if self._scene_combo_updating or self._session is None or not scene_name:
            return
        logger.info("Switching scene scene=%s", scene_name)
        try:
            changed = self._scene_residency.switch_scene(scene_name)
        except Exception:
            logger.exception("Failed to switch scene scene=%s", scene_name)
            self._set_status_override(
                f"Failed to switch to scene {scene_name}. See Log for details."
            )
            self._refresh_ui(full=True)
            return
        if not changed:
            return
        self._preferred_scene_name = scene_name
        self._status_override = None
        self._refresh_ui(full=True)
        self._request_viewport_redraw("switch_scene")

    def _on_property_edited(
        self,
        group_id: str,
        prop_index: int,
        value: object,
        expected_field: object,
        expected_title: object,
    ) -> None:
        if self._session is None:
            return
        logger.debug(
            "Applying property edit group=%s index=%s field=%s title=%s value=%r",
            group_id,
            prop_index,
            expected_field,
            expected_title,
            value,
        )
        try:
            update_mask = self._session.apply_property_input(
                group_id,
                prop_index,
                value,
                expected_field=None if expected_field is None else str(expected_field),
                expected_title=None if expected_title is None else str(expected_title),
            )
        except Exception:
            logger.exception(
                "Failed to apply property edit group=%s index=%s",
                group_id,
                prop_index,
            )
            self._set_status_override("Failed to apply property change. See Log for details.")
            self._refresh_ui(full=True)
            return
        self._status_override = None
        self._refresh_properties_panel(
            rebuild=bool(update_mask & UpdateMask.REBUILD_PROPERTY_SCHEMA)
        )
        self._refresh_summary()
        self._refresh_status_and_actions()
        self._request_viewport_redraw("apply_property_change")

    def save_image(self) -> None:
        if self._session is None:
            return
        current_size = self.viewport.host.size
        stored_size = self._settings.value("export_size")
        if isinstance(stored_size, (list, tuple)) and len(stored_size) == 2:
            try:
                current_size = int(stored_size[0]), int(stored_size[1])
            except Exception:
                current_size = self.viewport.host.size
        dialog = ExportImageDialog(
            width=current_size[0],
            height=current_size[1],
            parent=self,
        )
        if dialog.exec() != QtWidgets.QDialog.DialogCode.Accepted:
            return
        export_size = dialog.selected_size()
        filename, _filter = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Save Image",
            str(Path.home() / "render.png"),
            "PNG files (*.png);;JPEG files (*.jpg *.jpeg);;All files (*)",
        )
        if not filename:
            return
        logger.info("Saving image path=%s size=%s", filename, export_size)
        try:
            image = self._session.export_image(size=export_size)
            image.save(filename)
        except Exception as exc:
            logger.exception("Failed to save image path=%s size=%s", filename, export_size)
            self._set_status_override("Failed to save image. See Log for details.")
            QtWidgets.QMessageBox.warning(
                self,
                "Save Failed",
                f"Failed to save image to {filename!r}.\n\n{exc}",
            )
            return
        self._settings.setValue("export_size", list(export_size))
        self._status_override = None
        self.statusBar().showMessage(f"Saved image to {filename}", 5000)

    def _on_log_record_appended(self, levelno: int) -> None:
        if levelno < logging.WARNING:
            return
        self.log_dock.show()
        self.log_dock.raise_()
        self._sync_dock_area_actions()



import numpy as np
import matplotlib.pyplot as plt

from ..core import vector_to_rgb


def render_layer(
    director: np.ndarray,
    axis: int = 2,
    layer: int = 0,
    colormap=None,
    gridstep: float = 1,
    imagewidth_inch: float = 15,
    interpolation="none",
    length_unit="μm",
):
    if colormap is None:
        colormap = vector_to_rgb

    m = director[:, :, :, 0, :]

    if axis == 0:  # x-axis
        section = m[layer, :, :]
        axes = ["y", "z"]
    elif axis == 1:  # y-axis
        section = m[:, layer, :]
        axes = ["x", "z"]
    elif axis == 2:  # z-axis
        section = m[:, :, layer]
        axes = ["x", "y"]
    else:
        raise ValueError(f"Unsupported axis {axis}.")

    rgb = colormap(section)
    width, height, _ = section.shape
    aspectratio = width / height

    fig, ax = plt.subplots(figsize=(imagewidth_inch, imagewidth_inch / aspectratio))
    ax.imshow(
        rgb.transpose(1, 0, 2),
        origin="lower",
        extent=(0, gridstep * width, 0, gridstep * height),
        interpolation=interpolation,
        aspect="equal",
    )

    ax.set_xlabel(f"{axes[0]} [{length_unit}]")
    ax.set_ylabel(f"{axes[1]} [{length_unit}]")

    return fig, ax


#########################################################################################################################################


def plot_bloch_sphere(
    axis=2, colormap=None, size: int = 1024, imagewidth_inch: float = 5
):
    if colormap is None:
        colormap = vector_to_rgb

    # Bloch sphere plot.
    a = np.linspace(-1, 1, size)

    if axis == 0:  # x-axis
        y, z = np.meshgrid(a, a, indexing="ij")
        x = -np.sqrt(1 - y**2 - z**2)
        axes = ["y", "z"]
        yorigin = "lower"
    elif axis == 1:  # y-axis
        x, z = np.meshgrid(a, a, indexing="ij")
        y = -np.sqrt(1 - x**2 - z**2)
        axes = ["x", "z"]
        yorigin = "lower"
    elif axis == 2:  # z-axis
        x, y = np.meshgrid(a, a, indexing="ij")
        z = np.sqrt(1 - x**2 - y**2)
        axes = ["x", "y"]
        yorigin = "lower"
    else:
        raise ValueError(f"Unsupported axis {axis}.")

    xyz = np.stack([x, y, z], axis=-1)
    rgb = colormap(xyz)
    fig, ax = plt.subplots(figsize=(imagewidth_inch,) * 2, layout="compressed")
    ax.imshow(
        rgb.transpose(1, 0, 2),
        extent=[-1, 1, -1, 1] if yorigin == "lower" else [-1, 1, 1, -1],
        origin=yorigin,
    )
    ax.set_xlabel(axes[0])
    ax.set_ylabel(axes[1])
    return fig, ax
